/* calends.c - report useful things about a date

    calends [-c|e] [-l|b|r|f|n] [-svjtp|mqwya] [+[-]##] [yyyy-mm-dd]

    switch values:

        -h, -? : help

        -a : concatenate all offsets with holiday status
        -m : show the date offset within the month
        -q : offset within quarter
        -w : offset within week (Mon=0)
        -y : offset within year
        -s : set the dcl symbol CALEND_<yyyy-mm-dd> (VMS); simple '000'
        -v : verbose, show (don't set) the value for yyyy-mm-dd

        -c : use Catholic Gregorian conversion date (1582/10)
        -e : use English Gregorian conversion date (1752/09)

        -b : list combined holidays for the year of 'yyyy-mm-dd'
        -f : list Federal holidays for the year of 'yyyy-mm-dd'
        -l : list w/basis all holidays for the year 'yyyy-mm-dd'
        -n : list NYSE holidays for the year of 'yyyy-mm-dd'
        -r : list recognized holidays for the year of 'yyyy-mm-dd'

        -k : show this month, a vertical calendar

        -j : show Julian day values around 'yyyy-mm-dd'

        -t : test loop; show boundaries from Julian 0 to today

        -p : list lunar phases for the year of 'yyyy-mm-dd', with lunar holidays
        -x : -p, in .csv format

    date arithmetic:

        prefix a number (of days) with '+'; the number may be
        negative, which looks funny:

            calends +-365 (report on the date 365 days before today)

        or, provide 2 dates:

            > calends 2006-2-14 1950-5-19
            from 20060214 to 19500519 = -20360 days.

    build:

        vms:  cc calends; lin calends; calends:=$exe:calends

        gcc:  gcc -mno-cygwin -O2 -o calends.exe calends.c

        linux:  gcc [-DCALENDS_LIB=1] -O2 -o calends calends.c
          lib:  gcc [-DCALENDS_LIB=1] -O2 -o calends.o calends.c

          (defining CALENDS_LIB will compile only the functions, not the mainline;
          this can be linked with another mainline, such as a calendar)


*/
#define VERSION "2008309"
#define BUILD "Tue Nov  4 21:30:53 2008 309"
#define AUTHOR "Earl_Roosa"

#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <math.h>
#if VMS
#include <descrip.h>
#include <lib$routines.h>
#endif

int   ToJul(int y, int m, int d);
void  FromJul(int j, int *y, int *m, int *d);
int   LeapYear(int y);
int   EndOfMonth(int y, int m);
int   DayOfYear(int y, int m, int d);
int   fn_tojul(char *yyyymmdd);
char *fn_fromjul(int  j);
int   fn_oldcentury(int windowyear);
char *daytime(char *optional_caller_buffer);

#define eGDY1    1752
#define eGDM1    9
#define eGDD1    14
#define eGDFIX   11             // days dropped to correct the seasons
#define eGREGORYDAY1 2361222    // Sep 14, 1752

#define cGDY1    1582
#define cGDM1    10
#define cGDD1    15
#define cGDFIX   10             // days dropped
#define cGREGORYDAY1 2299161    // Oct 15, 1582

int GDY1        = cGDY1;
int GDM1        = cGDM1;
int GDD1        = cGDD1;
int GDFIX       = cGDFIX;
int GREGORYDAY1 = cGREGORYDAY1;

#define YEAR1   1721424         // Jan 1, 0001
#define WINDOWYEAR  70          // for mmddyy sliding window

#ifndef CALENDS_LIB

int   easter(int year,int *easter_month, int *easter_day);
int   roshhashanah(int year,int *month, int *day);
int   chanukah(int year,int *month, int *day);

int wksun = 1;                  // week starts on Sunday (EU likes Monday)

char *wday[]  = { "MO","TU","WE","TH","FR","SA","SU",NULL };
char *wkday[] = { "MON","TUE","WED","THU","FRI","SAT","SUN",NULL };

char *mnth[] = {
                    ""
                    "January",
                    "February",
                    "March",
                    "April",
                    "May",
                    "June",
                    "July",
                    "August",
                    "September",
                    "October",
                    "November",
                    "December",
                    ""
                };

typedef struct LUNAR { char *name; int(*fn)(int,int *,int *); int offset; } LUNAR;
LUNAR lunar[] = {
                    { "Easter"          ,easter         ,0    },
                    { "Good Friday"     ,easter         ,-2   },
                    { "Maundy Thursday" ,easter         ,-3   },
                    { "Palm Sunday"     ,easter         ,-7   },
                    { "Ash Wednesday"   ,easter         ,-46  },
                    { "Mardi Gras"      ,easter         ,-47  },
                    { "Ascension Day"   ,easter         ,39   },
                    { "Pentecost"       ,easter         ,49   },
                    { "Rosh Hashanah"   ,roshhashanah   ,0    },
                    { "Yom Kippur"      ,roshhashanah   ,9    },
                    { "Sukkot"          ,roshhashanah   ,14   },
                    { "Purim"           ,roshhashanah   ,-193 },
                    { "Passover"        ,roshhashanah   ,-163 },
                    { "Shavou'ot"       ,roshhashanah   ,-113 },
                    { "Chanukah"        ,chanukah       ,0 },
                    { NULL,NULL,0 } };

typedef struct HOLS { char *text; int m; int d; int y; int w; int j; } HOLS;
HOLS **hols = NULL;
HOLS **feds = NULL;
HOLS **nyse = NULL;
HOLS **luna = NULL;
HOLS **kals = NULL;

char *Federal[] = {
                    "    0101New Years Day",
                    " MO 0115ML King Day",
                    " MO 0215Presidents Day",
                    " MO 0525Memorial Day",
                    "    0704Independence Day",
                    " MO 0901Labor Day",
                    " MO 1008Columbus Day",
                    "    1111Armistice Day",
                    " TH 1122Thanksgiving Day",
                    "    1225Christmas Day",
                    NULL };

char *NY_SE[] = {
                    "Good Friday",
                    "    0101New Years Day",
                    " MO 0115ML King Day",
                    " MO 0215Presidents Day",
                    " MO 0525Memorial Day",
                    "    0704Independence Day",
                    " MO 0901Labor Day",
                    " TH 1122Thanksgiving Day",
                    "    1225Christmas Day",
                    NULL };

char *Holidays[] = {
                    "        ---Today--- ",
                    "        yyyymmdd (jjjjjjjj)     ",
                    "    0101New Years Day (FRB/NYSE)",
                    " .. 0106Twelfth Night",
                    " %410120Inauguration Day",
                    " MO 0115ML King Day (FRB/NYSE)",
                    " .. 0202Groundhog Day",
                    " .. 0212Lincoln",
                    " .. 0214Valentine's Day",
                    " MO 0215Presidents Day (FRB/NYSE)",
                    " .. 0221Washington",
                    " SU 0308Daylight Savings",
                    " .. 0317St Patrick's Day",
                    " .. 0401April Fool",
                    "Mardi Gras",
                    "Ash Wednesday",
                    "Palm Sunday",
                    "Maundy Thursday",
                    "Good Friday (NYSE)",
                    "Easter",
                    "Purim",
                    "Passover",
                    "Pentecost",
                    " SU 0508Mother's Day",
                    " MO 0525Memorial Day (FRB/NYSE)",
                    " SA 0515Armed Forces Day",
                    " .. 0614Flag Day",
                    " SU 0615Father's Day",
                    "    0704Independence Day (FRB/NYSE)",
                    " MO 0901Labor Day (FRB/NYSE)",
                    " .. 0911WTC Day",
                    "Rosh Hashanah",
                    "Yom Kippur",
                    "Sukkot",
                    " MO 1008Columbus Day (FRB)",
                    " .. 1031Halloween",
                    " .. 1101All Saints' Day",
                    " SU 1101Standard Time",
                    " TU 1102Election Day",
                    "    1111Armistice Day (FRB)",
                    " TH 1122Thanksgiving Day (FRB/NYSE)",
                    "Chanukah",
                    " .. 1201Advent",
                    " .. 1207Pearl Harbor Day",
                    " .. 1224Christmas Eve",
                    "    1225Christmas Day (FRB/NYSE)",
                    " .. 1225Christmas Day",
                    " .. 1226Boxing Day",
                    " .. 1226Kwanzaa",
                    NULL };

char **Kalends = Holidays;

char *Lunars[] = {
                    "Mardi Gras",
                    "Ash Wednesday",
                    "Palm Sunday",
                    "Maundy Thursday",
                    "Good Friday",
                    "Easter",
                    "Purim",
                    "Passover",
                    "Pentecost",
                    "Rosh Hashanah",
                    "Yom Kippur",
                    "Sukkot",
                    "Chanukah",
                    NULL };

int typ_def  = 1;
int typ_tst  = 0;
int typ_jul  = 0;
int typ_hol  = 0;
int typ_per  = 0;
int typ_set  = 0;
int typ_list = 0;
int typ_show = 0;
int typ_help = 0;
int typ_mon  = 0;
int typ_qua  = 0;
int typ_week = 0;
int typ_year = 0;
int typ_lun  = 0;
int typ_inc  = 0;
int typ_dff  = 0;
int typ_post = 0;
int typ_kal  = 0;
int typ_csv  = 0;

char **prnt = NULL;
int jan1 = 0;
char tday[40] = "";
char sday[40] = "";
char calendar[1024] = "";

void jyear(double td, long *yy, int *mm, int *dd);
void jhms(double j, int *h, int *m, int *s);
int jwday(double j);
double ucttoj(long year, int mon, int mday,  int hour, int min, int sec);
double phase(double  pdate, double  *pphase, double  *mage, double  *dist, double  *angdia, double  *sudist, double  *suangdia);
void phasehunt(double sdate, double phases[5]), fmt_phase_time(double utime, char *buf);

main(int argc, char *argv[])
{
    int i = 0;  // iter
    int k = 0;  // iter
    int r = 0;  // arg count
    int s = 0;  // status
    int a,z;    // first/last days of the period
    int h,n,m,d,y,mm,dd,yy;          // temporary dates
    long l;
    int j[] = { 0,0,0 };    // julian values, yesterday/today/tomorrow
    int v[] = { 0,0,0 };    // return values, 000-777
    char c = '\0';
    char *o = NULL;
    char x[] = { '*',')','@','(' };
    double f = 0.;
    double p[] = { 0., 0., 0., 0., 0., 0., 0., 0. };
    char aday[20] = "";
    char zday[20] = "";
    int  eday = 0;
    int  iday = 0;
    int  jday = 0;
    int  eom = 0;
    char tmp[1024] = "";
    char *mtext[44];
    char dtypkey[40] = "CALEND_LAST";
    char daykey[40] = "";
    char dayvalue[80] = "";
#if VMS
    $DESCRIPTOR(typdesc,dtypkey);
    $DESCRIPTOR(keydesc,daykey);
    $DESCRIPTOR(valdesc,dayvalue);
    int tbl = 2;
#endif
    void floatholiday(HOLS *h, int y);
    void findlunar(HOLS *h, int y);
    HOLS **readhols(char *ln, HOLS **h);
    int parsedate(char *b);
    void julian_conversion(char x);
    void showver();
    void showhlp(int typ_help);
    void addprnt(int j, char *s);
    HOLS **addhols(HOLS **c, HOLS **h);
    void readkalends(char *fn);

    for (r = 1; r < argc; r++) {
        //printf("ARG:  %s\n",argv[r]);
        typ_def = 0;
        if (argv[r][0] == '-') {
            for (i = 1; i < strlen(argv[r]); i++) {
                switch (argv[r][i]) {
                    case 'h': typ_help = 1; break;
                    case '?':
                    case 'H': typ_help = 2; break;
                    case 'a':
                              typ_show = 1;
                              typ_hol  = 1;
                              typ_mon  = 1;
                              typ_qua  = 1;
                              typ_week = 1;
                              typ_year = 1;
                              typ_per  = 1;
                              break;
                    case 'b': typ_post = 3; typ_help = 0; break;
                    case 'c': julian_conversion('c'); break;
                    case 'e': julian_conversion('e'); break;
                    case 'f': typ_post = 1; typ_help = 0; break;
                    case 'j': typ_jul  = 1; typ_hol  = 1; break;
                    case 'k': typ_kal  = 1; typ_help = 0; break;
                    case 'l': typ_list = 1; typ_hol  = 1; break;
                    case 'm': typ_mon  = 1; typ_per  = 1; break;
                    case 'n': typ_post = 2; typ_help = 0; break;
                    case 'p': typ_lun  = 1; typ_help = 0; break;
                    case 'q': typ_qua  = 1; typ_per  = 1; break;
                    case 'r': typ_post = 4; typ_help = 0; break;
                    case 's': typ_set  = 1; typ_hol  = 1; break;
                    case 't': typ_tst  = 1; break;
                    case 'v': typ_show = 1; typ_hol  = 1; break;
                    case 'w': typ_week = 1; typ_per  = 1; break;
                    case 'x': typ_csv  = 1; typ_lun  = 1; typ_help = 0; break;
                    case 'y': typ_year = 1; typ_per  = 1; break;
                    default:  if (isdigit(argv[r][1])) {
                                if (aday[0]) {
                                    typ_dff = 1;
                                    strncpy(zday,argv[r],19);
                                } else {
                                    strncpy(aday,argv[r],19);
                                }
                              }
                              break;
                }
            }
        } else if (argv[r][0] == '@') {
            strcpy(calendar,&argv[r][1]);
            readkalends(calendar);
        } else  if (argv[r][0] == '+') {
            typ_inc = 1;
            iday = atoi(&argv[r][1]);
        } else if (aday[0]) {
            typ_dff = 1;
            strncpy(zday,argv[r],19);
        } else {
            strncpy(aday,argv[r],19);
        }
    }
    if (!aday[0]) strncpy(aday,daytime(NULL),8);

    if ((jday=parsedate(aday)) < 0) {
        printf("!! date must be YYYY-MM-DD (%s)\n",aday);
        exit(2);
    }

    /*
    old calendar dates are not very useful to modern banking, but it's nice to
    be able to validate our julian day calculations.  julian days (named for julius
    scaliger, not julius caesar) begin on monday jan 1, 4713bce, and simply increment
    (at noon, UCT, if you're into astronomy).  that gives us some landmarks to inspect:
    day 0:  -4712/1/1 (4713,BCE); day 1721424:  0001/1/1; day 2299161:  gregorian
    conversion at the vatican, 1582/10/15; day 2361222:  gregorian conversion at
    whitehall, 1752/09/14; day 2,400,001:  smithsonian/vms day 0, 1858/11/17;
    there's no disagreement over any dates except the gregorian conversion, which
    was a dramatic late-renaissance demonstration of ideological denial of science.
    anyway, in the -t test loop, we report the conversion values of these dates,
    and every 100000th, to verify ToJul and FromJul.
    */
    if (typ_tst) {
        k = ToJul(-4712,1,1);
        while (k < jday) {
            FromJul(k,&y,&m,&d);
            if (k != (s = ToJul(y,m,d))) {
                printf("!! %d != %d, %d/%d/%d\n",k,s,y,m,d);
                exit(3);
            }
            if (((k % 100000) == 0)
             ||  (k ==  YEAR1)
             ||  (k == (YEAR1 - 1))
             ||  (k ==  GREGORYDAY1)
             ||  (k == (GREGORYDAY1 - 1))) {
                if (y<0) printf("%8d\t%05d/%02d/%02d\n",k,y,m,d);
                else     printf("%8d\t %04d/%02d/%02d\n",k,y,m,d);
            }
            k++;
        }
        FromJul(k,&y,&m,&d);
        printf("%8d\t %04d/%02d/%02d\n",k,y,m,d);
        printf("\n");
        showver();
        printf("\n");
        exit(0);
    }

    if (typ_inc) {
        jday += iday;
        strcpy(aday,fn_fromjul(jday));
        typ_help = 0;
    }
    if (typ_dff) {
        if ((iday=parsedate(zday)) < 0) {
            printf("!! date must be YYYY-MM-DD (%s)\n",zday);
            exit(2);
        }
        printf("from %s to %s = %d days.\n",aday,zday,iday-jday);
        typ_help = 0;
    }
    if ((typ_per) || (typ_hol)) {
        sprintf(daykey,"CALEND_%s",aday);
        typ_help = 0;
    }

    if (typ_help) showhlp(typ_help); // exit(0);

    if (typ_kal) {
        for (i = 0; i <= 43; i++) {
            if ((mtext[i] = calloc(1,1024)) == NULL) {
                printf("!! can't allocate 1024 bytes for line %d\n",i);
                exit(2);
            }
        }
        FromJul(jday,&y,&m,&d); jday -= (d-1); d = 1;
        eom = EndOfMonth(y,m);
        iday = ToJul(y,m,eom)+7;
        eday = ToJul(y,m,eom);
        r = (jday+wksun) % 7;
        for (i = 0; i < r; i++) strcpy(mtext[i],"  |");
        k = jday;
        while (k < ((jday-r)+43)) {
            FromJul(k,&yy,&mm,&dd);
            sprintf(mtext[i],"%2d|",dd);
            if (mm == m) eom = i;
            i++;
            k++;
        }
        f = ucttoj(y,m,1,0,0,0) - 30;
        while (f <= (double)eday+1) {
            phasehunt(f,p);
            for (i = 0; i < 4; i++) {
                if ((p[i] >= (double)jday) && (p[i] <= (double)eday+1)) {
                    jyear(p[i],&l,&n,&k); k += r-1;
                    jhms(p[i],&d,&n,&s);
                    sprintf(tmp,"%c %02d:%02d",x[i],d,n);
                    strcat(mtext[k],tmp);
                }
            }
            f = p[4] + 1;
        }
        strncpy(tday,daytime(NULL),8);
        strcat(tday,"--* Today *--");
        Kalends[0] = tday;
        sprintf(sday,"%8s--* %8s (%d) *--",aday,aday,fn_tojul(aday));
        Kalends[1] = sday;
        hols = readhols("KALENDS",kals);
        for (i = 0; (hols[i]) && (hols[i]->text); i++) {
            floatholiday(hols[i],y);
            if (hols[i]->j < jday) floatholiday(hols[i],y+1);
            if ((hols[i]->j >= jday) && (hols[i]->j <= iday)) {
                k = (hols[i]->j - jday) + r;
                if (isalpha(hols[i]->text[0])) o =  hols[i]->text;
                else                           o = &hols[i]->text[8];
                if (!strstr(mtext[k],o)) {
                    if (mtext[k][3]) strcat(mtext[k]," | "); else strcat(mtext[k]," ");
                    strcat(mtext[k],o);
                }
            }
        }
        s = ((jday-r) - (((ToJul(y,1,4) / 7) - 1) * 7)) / 7;
        printf("_____| %04d %s (%d-%d)\n",y,mnth[m-1],jday,iday);
        for (i = 0; i < 43; i++) {
            if (!(i%7)) {
                if (i) {
                    if (i <= eom) {
                        printf("__|__|:%02d\n",s);
                    } else {
                        printf("_____| %s\n",mnth[m%12]);
                    }
                    if (i == 42) break;
                    if ((i > 7) && (mtext[i][1] <= ' ')) break;
                }
                s++;
            }
            printf("%2s|%2s\n",wday[(i+6*wksun)%7],mtext[i]);
        }
    }

    if (typ_lun) {
        if (!prnt) prnt = calloc(366,sizeof(char *));
        FromJul(jday,&y,&m,&d);
        jan1 = ToJul(y,1,1);
        iday = ToJul(y+1,1,1);
        f = ucttoj(y,1,1,0,0,0) - 30;
        while (f <= (double)iday+1) {
            phasehunt(f,p);
            for (i = 0; i < 4; i++) {
                if ((p[i] >= (double)jan1) && (p[i] <= (double)iday+1)) {
                    jyear(p[i],&l,&m,&d);
                    jhms(p[i],&h,&n,&s);
                    //printf("%04d-%02d-%02d",l,m,d);
                    //printf("%c %02d:%02d GMT %12.8f",x[i],h,n,p[i]);
                    sprintf(tmp,"%c %02d:%02d",x[i],h,n);
                    addprnt(ToJul(l,m,d),tmp);
                }
            }
            f = p[4] + 1;
        }
        typ_post = 8;
    }

    if (typ_post) {
        if (!prnt) prnt = calloc(366,sizeof(char *));
        FromJul(jday,&y,&m,&d);
        jan1 = ToJul(y,1,1);
        for (k = 1; k < 16; k = (k << 1)) {
            hols = NULL;
            if ((typ_post & k) == 1) hols = readhols("KAL_FEDERAL",feds);
            if ((typ_post & k) == 2) hols = readhols("KAL_NYSE",nyse);
            if ((typ_post & k) == 4) hols = readhols("KALENDS",kals);
            if ((typ_post & k) == 8) hols = readhols("KAL_LUNARS",luna);
            if (hols) for (i = 0; (hols[i]) && (hols[i]->text); i++) {
                floatholiday(hols[i],y);
                if (hols[i]->j >= 0) {
                    FromJul(hols[i]->j,&y,&m,&d);
                    if (isalpha(hols[i]->text[0])) o =  hols[i]->text;
                    else                           o = &hols[i]->text[8];
                    //printf("%04d-%02d-%02d %s\n",y,m,d,o);
                    addprnt(hols[i]->j,o);
                }
            }
        }
        typ_post = 0;
    }

    if ((typ_def) || (typ_hol)) {
        j[0] = fn_tojul(aday) - 1;
        for (i = 0; i < 3; i++) {
            j[i] = j[0] + i;
            v[i] = ((j[i] % 7) > 4);
        }
        if (typ_jul) {
            FromJul(j[0],&y,&m,&d);
            printf("Julian day: %d (%04d.%02d.%02d)\n",j[0],y,m,d);
            FromJul(j[1],&y,&m,&d);
            printf("Julian day: %d (%04d.%02d.%02d)",j[1],y,m,d);
            printf(" Leap year: %d\n",LeapYear(y));
            FromJul(j[2],&y,&m,&d);
            printf("Julian day: %d (%04d.%02d.%02d)\n",j[2],y,m,d);
            //exit(0);
        }
        if (typ_list) {
            FromJul(j[1],&y,&m,&d);
            printf("Julian day: %d (%d-%d-%d / %d)",j[1],y,m,d,DayOfYear(y,m,d));
            printf(" Leap year: %d\n",LeapYear(y));
        }
        feds = readhols("KAL_FEDERAL",feds);
        for (i = 0; (feds[i]) && (feds[i]->text); i++) {
            for (k = 0; k < 3; k++) {
                FromJul(j[k],&y,&m,&d);
                floatholiday(feds[i],y);
                if (j[k] == feds[i]->j) v[k] |= 2;
                if ((k==1) && (typ_list)) {
                  //printf("BaseDate DOW Floated  |Text         (%s %p)\n",fn,fp);
                    printf(" %02d-%02d   %2d  ",feds[i]->m,feds[i]->d,feds[i]->w);
                    printf("%s |%s\n",fn_fromjul(feds[i]->j),feds[i]->text);
                }
            }
        }
        nyse = readhols("KAL_NYSE",nyse);
        for (i = 0; (nyse[i]) && (nyse[i]->text); i++) {
            for (k = 0; k < 3; k++) {
                FromJul(j[k],&y,&m,&d);
                floatholiday(nyse[i],y);
                if (j[k] == nyse[i]->j) v[k] |= 4;
                if ((k==1) && (typ_list)) {
                  //printf("BaseDate DOW Floated  |Text         (%s %p)\n",fn,fp);
                    printf(" %02d-%02d   %2d  ",nyse[i]->m,nyse[i]->d,nyse[i]->w);
                    printf("%s |%s\n",fn_fromjul(nyse[i]->j),nyse[i]->text);
                }
            }
        }
        sprintf(dayvalue,"%d%d%d",v[0],v[1],v[2]);
    }
    if (typ_per) {
        k = fn_tojul(aday);
        FromJul(k,&y,&m,&d); mm=r=m; yy=s=y;
        if (typ_week) {
            a = k % 7;
            z = 7 - a;
            c = 'W';
            sprintf(tmp,"_%03d_%03d",a,z); strcat(dayvalue,tmp);
        }
        if (typ_mon) {
            a = k - ToJul(y,m,1);
            if (mm == 12) { s++; r = 1; } else r++;
            z = ToJul(s,r,1) - k;
            c = 'M';
            sprintf(tmp,"_%03d_%03d",a,z); strcat(dayvalue,tmp);
        }
        if (typ_qua) {
            r = ((mm-1) / 3) * 3 + 1;
            a = k - ToJul(yy,r,1);
            if (r == 10) { yy++; r = 1; } else r+=3;
            z = ToJul(yy,r,1) - k;
            c = 'Q';
            sprintf(tmp,"_%03d_%03d",a,z); strcat(dayvalue,tmp);
        }
        if (typ_year) {
            a = k - ToJul(y,1,1);
            z = ToJul(y+1,1,1) - k;
            c = 'Y';
            sprintf(tmp,"_%03d_%03d",a,z); strcat(dayvalue,tmp);
        }
    }
    if ((typ_per) || (typ_hol)) {
        if (typ_show)
            printf("%s=%s\n",daykey,dayvalue);
        else
            printf("%s\n",dayvalue);
#if VMS
        if (typ_set) {
            keydesc.dsc$w_length = strlen(daykey);
            valdesc.dsc$w_length = strlen(dayvalue);
            typdesc.dsc$w_length = strlen(dtypkey);
            s = lib$set_symbol(&keydesc,&valdesc,&tbl);
            s = lib$set_symbol(&typdesc,&keydesc,&tbl);
        }
#endif
    }

    if (prnt) {
        if (typ_csv) {
            printf("CSV datebook: Category, Private, Description, Note, Event, Begin, End, Alarm, ");
            printf("Advance, Advance Units, Repeat Type, Repeat Forever, Repeat End, Repeat Frequency, ");
            printf("Repeat Day, Repeat Days, Week Start, Number of Exceptions, Exceptions\n");
        }
        for (i = 0; i < 366; i++) {
            if (prnt[i]) {
                FromJul(jan1+i,&y,&m,&d);
                if (typ_csv) {
                    printf("\"\",\"0\",\"%s\",\"\",\"1\",",prnt[i]);
                    printf("\"%04d %02d %02d  00:00\",\"%04d %02d %02d  00:00\",",y,m,d,y,m,d);
                    printf("\"0\",\"0\",\"0\",\"0\",\"1\",\"136560132 134910321 -1216122892\",");
                    printf("\"0\",\"0\",\"0000000\",\"0\",\"0\",\"\"\n");
                } else {
                    printf("%04d-%02d-%02d\t%s\n",y,m,d,prnt[i]);
                }
            }
        }
    }

    if (typ_def) printf("%s:  %s\n",daytime(NULL),dayvalue);

    exit(0);
}

void julian_conversion(char x)
{
    if (x == 'c') {
        GDY1        = cGDY1;
        GDM1        = cGDM1;
        GDD1        = cGDD1;
        GDFIX       = cGDFIX;
        GREGORYDAY1 = cGREGORYDAY1;
    } else {
        GDY1        = eGDY1;
        GDM1        = eGDM1;
        GDD1        = eGDD1;
        GDFIX       = eGDFIX;
        GREGORYDAY1 = eGREGORYDAY1;
    }
}

void readkalends(char *fn) {
    int i = 0;
    int n = 0;
    FILE *fp = NULL;
    char b[1024] = "";

    if (Kalends == Holidays) {
        Kalends = NULL;
        while (Holidays[i]) {
            if (i >= n) {
                n = i + 12;
                Kalends = realloc(Kalends,(n+2)*sizeof(char **));
            }
            Kalends[i] = calloc(1,strlen(Holidays[i])+1);
            strcpy(Kalends[i],Holidays[i]);
            //printf("k %3d %s\n",i,Kalends[i]);
            i++;
        }
        Kalends[i] = NULL;
    }
    while ((Kalends) && (Kalends[i])) {
        //printf("k %3d %s\n",i,Kalends[i]);
        i++;
    }
    if (fp = fopen(fn,"r")) {
        while (fgets(b,1022,fp)) {
            while ((strlen(b)) && (b[strlen(b)-1] <= ' ')) b[strlen(b)-1] = '\0';
            if (strlen(b)) {
                if (i >= n) {
                    n = i + 12;
                    Kalends = realloc(Kalends,(n+2)*sizeof(char **));
                }
                Kalends[i] = calloc(1,strlen(b)+1);
                strcpy(Kalends[i],b);
                //printf("k %3d %s\n",i,Kalends[i]);
                i++;
            }
        }
        Kalends[i] = NULL;
    }
}

void addprnt(int j, char *s) {
    if (prnt[j-jan1]) {
        prnt[j-jan1] = realloc(prnt[j-jan1],strlen(prnt[j-jan1])+strlen(s)+4);
        if (!strstr(prnt[j-jan1],s)) {
            strcat(prnt[j-jan1],"\t");
            strcat(prnt[j-jan1],s);
        }
    } else {
        prnt[j-jan1] = calloc(1,strlen(s)+4);
        strcpy(prnt[j-jan1],s);
    }
}

HOLS **addhols(HOLS **h, HOLS **c)
{
    int i = 0;
    int hh = 0;
    int cc = 0;

    if (!c) return(h);
    while ((c) && (c[cc]) && (c[cc]->text)) {
        //printf("c %d %s\n",cc,c[cc]->text);
        cc++;
    }
    while ((h) && (h[hh]) && (h[hh]->text)) {
        //printf("h %d %s\n",hh,h[hh]->text);
        hh++;
    }
    h = realloc(h,(cc+hh+2)*sizeof(HOLS *));
    while (c[i] && c[i]->text) {
        h[hh] = calloc(1,sizeof(HOLS));
        h[hh]->text = calloc(1,strlen(c[i]->text)+1);
        strcpy(h[hh]->text,c[i]->text);
        //printf("h %d %s\n",hh,h[hh]->text);
        h[hh]->m = c[i]->m;
        h[hh]->d = c[i]->d;
        h[hh]->y = c[i]->y;
        h[hh]->w = c[i]->w;
        h[hh]->j = c[i]->j;
        hh++;
        i++;
    }
    return(h);
}

HOLS **readhols(char *ln, HOLS **h)
{
    int i = 0;
    int j = 0;
    int n = 0;
    char b[256] = "";
    char *nxthol(char *ln, char *b, int *i);

    while (nxthol(ln,b,&i)) {
        if (i >= n) {
            n += 12;
            h = realloc(h,n*sizeof(HOLS *));
        }
        h[i] = calloc(1,sizeof(HOLS));
        h[i]->text = calloc(1,strlen(b)+1);
        strcpy(h[i]->text,b);
        i++;
    }
    h[i] = calloc(1,sizeof(HOLS));
    for (i = 0; (h[i]) && (h[i]->text); i++) {
        if (h[i]->text[0] == ' ') {
            memcpy(b,&h[i]->text[1],2); b[2] = '\0';
            if (h[i]->text[1] > '.') {
                for (n = 0; (wday[n]) && (strcmp(b,wday[n])); n++) ;
                if (wday[n]) h[i]->w = n;
                else         h[i]->w = -1;
            } else if (h[i]->text[1] == '%') {
                h[i]->w = -4;
                h[i]->y = h[i]->text[2] & 0x0f;
            } else if (h[i]->text[1] == '.') {
                h[i]->w = -2;
            } else {
                h[i]->w = -1;
            }
            h[i]->m = (h[i]->text[4] & 15) * 10 + (h[i]->text[5] & 15);
            h[i]->d = (h[i]->text[6] & 15) * 10 + (h[i]->text[7] & 15);
        } else if (isdigit(h[i]->text[0])) {
            h[i]->w = -5;
            h[i]->y = 0;
            for (j = 0; j < 4; j++) {
                h[i]->y *= 10;
                h[i]->y += (h[i]->text[j] & 15);
            }
            h[i]->m = (h[i]->text[4] & 15) * 10 + (h[i]->text[5] & 15);
            h[i]->d = (h[i]->text[6] & 15) * 10 + (h[i]->text[7] & 15);
        } else {
            h[i]->w = -3;
        }
        //printf("rh %s %d %d %d %d %d\n",h[i]->text,h[i]->y,h[i]->m,h[i]->d,h[i]->w,h[i]->j);
    }
    return(h);
}

char *nxthol(char *ln, char *b, int *n)
{
    int i = 0;
    int l = 0;
    int s = 0;
    static FILE *fp = NULL;
    char fn[128] = "";
    char *e = NULL;

    if (*n == 0) {
        if (e = getenv(ln)) {
            strcpy(fn,e);
        } else {
            strcpy(fn,ln);
        }
        fp = fopen(fn,"r");
        if (typ_list && !typ_post) {
            printf("BaseDate DOW Floated  |Text         (%s",fn);
            if (fp) printf(" %p",fp);
            printf(")\n");
        }
    }
    do {
        if (fp) {
            if (!(fgets(b,255,fp))) {
                fclose(fp);
                return(NULL);
            }
            while ((strlen(b)) && (b[strlen(b)-1] <= ' ')) b[strlen(b)-1] = '\0';
        } else {
            if (!strcmp(ln,"KAL_HOLIDAYS")) e = Holidays[*n];
            if (!strcmp(ln,"KAL_FEDERAL"))  e = Federal[*n];
            if (!strcmp(ln,"KAL_NYSE"))     e = NY_SE[*n];
            if (!strcmp(ln,"KAL_LUNARS"))   e = Lunars[*n];
            if (!strcmp(ln,"KALENDS"))      e = Kalends[*n];
            if ((e) && (e[0]))
                strcpy(b,e);
            else
                return(NULL);
        }
        if (b[0] > '@') {
            for (i = 0; (lunar[i].name); i++) {
                l = (strlen(lunar[i].name) < strlen(b) ? strlen(lunar[i].name) : strlen(b));
                if (!strncmp(b,lunar[i].name,l)) s = 1;
            }
            if (!s) *n++;
        } else {
            s = 1;
        }
    } while (s == 0);
    if (b[0] < ' ') return(NULL);
    return(b);
}

void findlunar(HOLS *h, int y)
{
    int     i = 0;
    int     j = 0;
    int     l = 0;
    int     m = 0;
    int     d = 0;
    void floatholiday(HOLS *h, int y);

    if (h->w != -3) {
        floatholiday(h,y);
        return;
    }
    h->y = y;
    for (i = 0; (lunar[i].name); i++) {
        l = (strlen(lunar[i].name) < strlen(h->text) ? strlen(lunar[i].name) : strlen(h->text));
        if (!strncmp(h->text,lunar[i].name,l)) {
            lunar[i].fn(y,&m,&d);
            h->m = m;
            h->d = d;
            h->j = ToJul(y,m,d) + lunar[i].offset;
            break;
        }
    }
}

/*
    there are several holiday types to float:
        -1 - New Year's (e.g.) move it to Monday only if it falls on Sunday
        -2 - don't move it, no matter when it falls
        -3 - a lunar holiday, go to the Easter/RoshHashana functions
        -4 - not annual, Inauguration Day (e.g.) %41 = year 1 of every 4th
        -5 - an actual, complete date, with a year (like a log entry)
        0...6 - float forward to this day of the week (0=Mon)
*/

void floatholiday(HOLS *h, int y)
{
    int     i = 0;
    void findlunar(HOLS *h, int y);

    if (h->w == -3) {
        findlunar(h,y);
        return;
    }
    if (h->w == -4) {
        i = h->text[3] & 0x0f;
        if ((y % h->y) != i) {
            h->j = -1;
            return;
        }
        h->w = -1;
    }
    //printf("fh1 %s %d %d %d %d %d\n",h->text,h->y,h->m,h->d,h->w,h->j);
    if (h->w == -5) {
        if (y != h->y) {
            h->j = -1;
            return;
        }
        h->w = -2;
    }
    h->y = y;
    h->j = ToJul(y,h->m,h->d);
    //printf("fh2 %s %d %d %d %d %d\n",h->text,h->y,h->m,h->d,h->w,h->j);
    if (h->w == -2) return;
    i = (h->j % 7);
    if (h->w == -1) {
        if (i == 6) h->j++;
        return;
    }
    if (i != h->w) {
        h->j += 7 - (((i + 7) - h->w) % 7);
    }
    //printf("fh3 %s %d %d %d %d %d\n",h->text,h->y,h->m,h->d,h->w,h->j);
}

int easter(int year,int *easter_month, int *easter_day)
{
    int a,b,c,e,g,h,i,k,u,x,z;
    div_t f;

    /*
    **  Gauss' famous algorithm (I don't know how or why it works,
    **  so there's no commenting)
    */

    a = year % 19;
    f = div(year,100);
    b = f.quot;
    c = f.rem;
    f = div(b,4);
    z = f.quot;
    e = f.rem;
    f = div((8*b + 13),25);
    g = f.quot;
    f = div((19*a + b - z - g + 15),30);
    h = f.rem;
    f = div((a + 11*h),319);
    u = f.quot;
    f = div(c,4);
    i = f.quot;
    k = f.rem;
    f = div((2*e + 2*i - k - h + u + 32),7);
    x = f.rem;
    f = div((h-u+x+90),25);
    *easter_month = f.quot;
    f = div((h-u+x + *easter_month +19),32);
    *easter_day = f.rem;
    return((int)ToJul(year,*easter_month,*easter_day));
}

#include <math.h>

int roshhashanah(int year,int *month, int *day)
{
    double G = 0.;
    double N = 0.;
    double c = 0.;
    double y = 0.;
    double f = 0.;
    double r = 0.;
    int j = 0;
    int d = 0;
    int m = 0;
    int w = 0;
    double Remainder(double x, double y);

    y = year;
    c = floor(y / 100.);
    G = Remainder(y,19) + 1.;
    r = Remainder(12*G,19);
    N = ((floor(y/100.)-floor(y/400.))-2.)
        + (765433./492480.)*r
        + Remainder(y,4.)/4.
        - (313*(y)+89081)/98496.;
    f = N - floor(N);
    j = ToJul(year,8,31) + (int)N;
    w = j % 7;
    if ((w == 2) || (w == 4) || (w == 6)) j++;
    else
    if ((w == 0) && (r > 11.) && (f >= 0.898)) j++;
    else
    if ((w == 1) && (r >  6.) && (f >= 0.633)) j+=2;
    FromJul(j,&w,month,day);
    return(j);
}

/* chanukah turns out to be an easy calculation, once you have rosh ha-shanah;
    in the metonic cycle, there are years of 353, 354, 355, 383, 384, and 385
    days; in the 'complete' years, (355 or 385), cheshvan has 30 days (as does
    kislev, but that doesn't matter), so the time from rosh ha-shanah to 
    chanukah is 29+30+25 = 84 days, instead of 29+29+25 = 83.
*/
int chanukah(int year,int *month, int *day)
{
    int r1 = 0;
    int r2 = 0;
    int y  = year;

    r2 = roshhashanah(year+1,month,day);
    r1 = roshhashanah(year  ,month,day);
    if (((r2-r1) % 5) == 0) r1 += 84; else r1 += 83;
    FromJul(r1,&y,month,day);
    return(r1);
}

double Remainder(double x, double y)
{
    double r;
    r = (x - (y * floor(x / y)));
    return(r);
}

/* lunar phases

    This program was written by:

        John Walker
                http://www.fourmilab.ch/

    This  program is in the public domain: "Do what thou wilt shall be
    the whole of the law".  I'd appreciate  receiving  any  bug  fixes
    and/or  enhancements, which I'll incorporate in future versions of
    the program.  Please leave the  original  attribution  information
    intact so that credit and blame may be properly apportioned.

    ---------
    Perhaps my glomming useful library functions for incorporation in
    my own calendar tool is beyond Mr Walker's intentions; I hope not.
    The code is way beyond my capability, even if I had the motivation
    to do the research, which I don't.  So I thank John Walker for
    sharing. -e.roosa
*/

/*  Astronomical constants  */

#define epoch       2444238.5      /* 1980 January 0.0 */

/*  Constants defining the Sun's apparent orbit  */

#define elonge      278.833540     /* Ecliptic longitude of the Sun
                                      at epoch 1980.0 */
#define elongp      282.596403     /* Ecliptic longitude of the Sun at
                                      perigee */
#define eccent      0.016718       /* Eccentricity of Earth's orbit */
#define sunsmax     1.495985e8     /* Semi-major axis of Earth's orbit, km */
#define sunangsiz   0.533128       /* Sun's angular size, degrees, at
                                      semi-major axis distance */

/*  Elements of the Moon's orbit, epoch 1980.0  */

#define mmlong      64.975464      /* Moon's mean longitude at the epoch */
#define mmlongp     349.383063     /* Mean longitude of the perigee at the
                                      epoch */
#define mlnode      151.950429     /* Mean longitude of the node at the
                                      epoch */
#define minc        5.145396       /* Inclination of the Moon's orbit */
#define mecc        0.054900       /* Eccentricity of the Moon's orbit */
#define mangsiz     0.5181         /* Moon's angular size at distance a
                                      from Earth */
#define msmax       384401.0       /* Semi-major axis of Moon's orbit in km */
#define mparallax   0.9507         /* Parallax at distance a from Earth */
#define synmonth    29.53058868    /* Synodic month (new Moon to new Moon) */
#define lunatbase   2423436.0      /* Base date for E. W. Brown's numbered
                                      series of lunations (1923 January 16) */

/*  Properties of the Earth  */

#define earthrad    6378.16        /* Radius of Earth in kilometres */

#define PI 3.14159265358979323846  /* Assume not near black hole nor in
                                      Tennessee */

/*  Handy mathematical functions  */

#define sgn(x) (((x) < 0) ? -1 : ((x) > 0 ? 1 : 0))       /* Extract sign */
#define Abs(x) ((x) < 0 ? (-(x)) : (x))                   /* Absolute val */
#define fixangle(a) ((a) - 360.0 * (floor((a) / 360.0)))  /* Fix angle    */
#define torad(d) ((d) * (PI / 180.0))                     /* Deg->Rad     */
#define todeg(d) ((d) * (180.0 / PI))                     /* Rad->Deg     */
#define dsin(x) (sin(torad((x))))                         /* Sin from deg */
#define dcos(x) (cos(torad((x))))                         /* Cos from deg */

int testmode = 0;                   /* Rapid warp through time for debugging */
int runmode = 1;                    /* Update time if true */
double faketime = 0.0;              /* Time increment for test mode */
static int color_mode = 0;          /* Indicates color/mono mode */
static double nptime = 0.0;         /* Next new moon time */

/*  UCTTOJ  --  Convert GMT date and time to astronomical
                                Julian time (i.e. Julian date plus day fraction,
                                expressed as a double).  */

double ucttoj(long year, int mon, int mday,  int hour, int min, int sec)
{

        // Algorithm as given in Meeus, Astronomical Algorithms, Chapter 7, page 61

        int a, b, m;
        long y;

#ifdef PARANOID
        assert(mon  >= 0 && mon  < 12);
        assert(mday >  0 && mday < 32);
        assert(hour >= 0 && hour < 24);
        assert(min  >= 0 && min  < 60);
        assert(sec  >= 0 && sec  < 60);
#endif

        m = mon + 1;
        y = year;

        if (m <= 2) {
                y--;
                m += 12;
        }

        /* Determine whether date is in Julian or Gregorian calendar based on
           canonical date of calendar reform. */

        if ((year < GDY1) || ((year == GDY1) && ((mon < (GDM1-1)) || (mon == (GDM1-1) && mday < GDD1)))) {
                b = 0;
        } else {
                a = ((int) (y / 100));
                b = 2 - a + (a / 4);
        }

        return (((long) (365.25 * (y + 4716))) + ((int) (30.6001 * (m + 1))) +
                                mday + b - 1524.5) +
                        ((sec + 60L * (min + 60L * hour)) / 86400.0);
}

/*  JYEAR  --  Convert  Julian  date  to  year,  month, day, which are
                           returned via integer pointers to integers (note that year is a long).  */

void jyear(double td, long *yy, int *mm, int *dd)
{
        double z, f, a, alpha, b, c, d, e;

        td += 0.5;
        z = floor(td);
        f = td - z;

        if (z < (double)GREGORYDAY1) {
                a = z;
        } else {
                alpha = floor((z - 1867216.25) / 36524.25);
                a = z + 1 + alpha - floor(alpha / 4);
        }

        b = a + 1524;
        c = floor((b - 122.1) / 365.25);
        d = floor(365.25 * c);
        e = floor((b - d) / 30.6001);

        *dd = (int) (b - d - floor(30.6001 * e) + f);
        *mm = (int) ((e < 14) ? (e - 1) : (e - 13));
        *yy = (long) ((*mm > 2) ? (c - 4716) : (c - 4715));
}

/*  JHMS  --  Convert Julian time to hour, minutes, and seconds.  */

void jhms(double j, int *h, int *m, int *s)
{
    long ij;

    j += 0.5;                         /* Astronomical to civil */
    ij = (long) (((j - floor(j)) * 86400.0) + 0.5);  // Round to nearest second
    *h = (int) (ij / 3600L);
    *m = (int) ((ij / 60L) % 60L);
    *s = (int) (ij % 60L);
}

/*      JWDAY  --  Determine day of the week for a given Julian day.  */

int jwday(double j)
{
        return ((int) (j + 1.5)) % 7;
}

/*  MEANPHASE  --  Calculates  time  of  the mean new Moon for a given
                   base date.  This argument K to this function is the
                   precomputed synodic month index, given by:

                          K = (year - 1900) * 12.3685

                   where year is expressed as a year and fractional year.  */

static double meanphase(double sdate, double k)
{
    double t, t2, t3, nt1;

    /* Time in Julian centuries from 1900 January 0.5 */
    t = (sdate - 2415020.0) / 36525;
    t2 = t * t;                       /* Square for frequent use */
    t3 = t2 * t;                      /* Cube for frequent use */

    nt1 = 2415020.75933 + synmonth * k
            + 0.0001178 * t2
            - 0.000000155 * t3
            + 0.00033 * dsin(166.56 + 132.87 * t - 0.009173 * t2);

    return nt1;
}

/*  TRUEPHASE  --  Given a K value used to determine the mean phase of
                   the new moon, and a phase selector (0.0, 0.25, 0.5,
                   0.75), obtain the true, corrected phase time.  */

static double truephase(double k, double phase)
{
    double t, t2, t3, pt, m, mprime, f;
    int apcor = 0;

    k += phase;                       /* Add phase to new moon time */
    t = k / 1236.85;                  /* Time in Julian centuries from
                                         1900 January 0.5 */
    t2 = t * t;                       /* Square for frequent use */
    t3 = t2 * t;                      /* Cube for frequent use */
    pt = 2415020.75933                /* Mean time of phase */
         + synmonth * k
         + 0.0001178 * t2
         - 0.000000155 * t3
         + 0.00033 * dsin(166.56 + 132.87 * t - 0.009173 * t2);

    m = 359.2242                      /* Sun's mean anomaly */
        + 29.10535608 * k
        - 0.0000333 * t2
        - 0.00000347 * t3;
    mprime = 306.0253                 /* Moon's mean anomaly */
        + 385.81691806 * k
        + 0.0107306 * t2
        + 0.00001236 * t3;
    f = 21.2964                       /* Moon's argument of latitude */
        + 390.67050646 * k
        - 0.0016528 * t2
        - 0.00000239 * t3;
    if ((phase < 0.01) || (Abs(phase - 0.5) < 0.01)) {

       /* Corrections for New and Full Moon */

       pt +=     (0.1734 - 0.000393 * t) * dsin(m)
                + 0.0021 * dsin(2 * m)
                - 0.4068 * dsin(mprime)
                + 0.0161 * dsin(2 * mprime)
                - 0.0004 * dsin(3 * mprime)
                + 0.0104 * dsin(2 * f)
                - 0.0051 * dsin(m + mprime)
                - 0.0074 * dsin(m - mprime)
                + 0.0004 * dsin(2 * f + m)
                - 0.0004 * dsin(2 * f - m)
                - 0.0006 * dsin(2 * f + mprime)
                + 0.0010 * dsin(2 * f - mprime)
                + 0.0005 * dsin(m + 2 * mprime);
       apcor = 1;
    } else if ((Abs(phase - 0.25) < 0.01 || (Abs(phase - 0.75) < 0.01))) {
       pt +=     (0.1721 - 0.0004 * t) * dsin(m)
                + 0.0021 * dsin(2 * m)
                - 0.6280 * dsin(mprime)
                + 0.0089 * dsin(2 * mprime)
                - 0.0004 * dsin(3 * mprime)
                + 0.0079 * dsin(2 * f)
                - 0.0119 * dsin(m + mprime)
                - 0.0047 * dsin(m - mprime)
                + 0.0003 * dsin(2 * f + m)
                - 0.0004 * dsin(2 * f - m)
                - 0.0006 * dsin(2 * f + mprime)
                + 0.0021 * dsin(2 * f - mprime)
                + 0.0003 * dsin(m + 2 * mprime)
                + 0.0004 * dsin(m - 2 * mprime)
                - 0.0003 * dsin(2 * m + mprime);
       if (phase < 0.5)
          /* First quarter correction */
          pt += 0.0028 - 0.0004 * dcos(m) + 0.0003 * dcos(mprime);
       else
          /* Last quarter correction */
          pt += -0.0028 + 0.0004 * dcos(m) - 0.0003 * dcos(mprime);
       apcor = 1;
    }
    return pt;
}

/*   PHASEHUNT  --  Find time of phases of the moon which surround the
                    current date.  Five phases are found, starting and
                    ending with the new moons which bound the  current
                    lunation.  */

void phasehunt(double sdate, double phases[6])
{
    double adate, k1, k2, nt1, nt2;
    long yy;
    int mm, dd;

    adate = sdate - 45;

    jyear(adate, &yy, &mm, &dd);
    k1 = floor((yy + ((mm - 1) * (1.0 / 12.0)) - 1900) * 12.3685);

    adate = nt1 = meanphase(adate, k1);
    while (1) {
        adate += synmonth;
        k2 = k1 + 1;
        nt2 = meanphase(adate, k2);
        if (nt1 <= sdate && nt2 > sdate)
            break;
        nt1 = nt2;
        k1 = k2;
    }
    phases[0] = truephase(k1, 0.0);
    phases[1] = truephase(k1, 0.25);
    phases[2] = truephase(k1, 0.5);
    phases[3] = truephase(k1, 0.75);
    phases[4] = truephase(k2, 0.0);
    phases[5] = truephase(k2, 0.25);
}
#endif

int parsedate(char *b)
{
    int  i = 0;
    int  j = 0;
    int  k = 0;
    int  l = 0;
    int  n[] = { 0,0,0 };
    int  i2 = 0;
    int  i3 = 0;
    int  im = 0;
    int  id = 0;
    int  iy = 0;
    int  ymd = 0;
    char tmp[20] = "";
    char *d[20] = { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL
                   ,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };
    struct tm  *now;
    time_t      secs;

    if (strlen(b) > 19) return(-1);
    k = j = i = 0;
    ymd = (b[0] == '-');
    while ((b[k]) && (i < 3)) {
        if (((j == 4) && (isdigit(b[0])))
        ||  ((j == 2) && (i))
        ||  ((!isdigit(b[k])) && (k))) {
            i++;
            j = 0;
        }
        if (isdigit(b[k]) || ((k == 0) && (b[k] == '-'))) {
            if (j==0) d[i] = calloc(1,6);
            d[i][j++] = b[k];
        }
        //printf("%d %d %d %d\n",i,j,k,d[i]);
        i += (!b[++k]);
    }
    //printf("%d %d %d\n",i,j,k);
    for (j = 0; (j < i) && (d[j]); j++) n[j] = atoi(d[j]);
    if (ymd) {
        iy = n[0]; im = n[1]; id = n[2];
    } else {
        if      (n[0] > 31) { iy = n[0]; im = n[1]; id = n[2]; }
        else if (n[0] > 12) { id = n[0]; im = n[1]; iy = n[2]; }
        else                { im = n[0]; id = n[1]; iy = n[2]; }
    }
    if (im == 0) {
        time(&secs);
        now = localtime(&secs);
        iy = now->tm_year + 1900;
        im = now->tm_mon  + 1;
        id = now->tm_mday;
    }
    j = ToJul(iy,im,id);
    //printf("%d %d %d %d ,%d #%d\n",j,im,id,iy,LeapYear(iy),DayOfYear(iy,im,id));
    //FromJul(j,&iy,&im,&id);
    //printf("%d %d %d %d\n",j,im,id,iy);
    //exit(0);
    if (j >= 0) strcpy(b,fn_fromjul(j));
    return(j);
}

int EndOfMonth(int y, int m)
{
    int d = 30;

    d += ((m < 8)  && ((m & 1) == 1));
    d += ((m > 7)  && ((m & 1) == 0));
    d -=  (m == 2);
    d -= ((m == 2) && (LeapYear(y) == 0));
    return(d);
}

int DayOfYear(int y, int m, int d)
{
    int i = 1;
    int n = d;

    for (i = 1; i < m; i++) {
        n += EndOfMonth(y,i);
    }
    return(n);
}

/* convert a 3-int date to a Julian
    first check that the date is valid
*/
int ToJul(int y, int m ,int d)
{
    int    j = 0;
    int    n = 0;

    if ((m < 1)     || (m > 12)
    ||  (d < 1)     || (d > 31)
    ||  ((m == 2)   && (d > (28 + LeapYear(y))))
    ||  ((m <  8)   && !(m & 1) && (d > 30))
    ||  ((m >  7)   &&  (m & 1) && (d > 30))
    ||  ((y == GDY1) && (m == GDM1) && ((d >= (GDD1 - GDFIX)) && (d < GDD1)))
    )
        return(-1);

    if (y < GDY1) {
        j = ((y + 4712) / 4) * 1461;
        if (n = ((y + 4712) % 4) * 365) j += n + 1;
        j += DayOfYear(y,m,d) - 1;
    } else {
        n = (m - 14) / 12;
        if ((j = d - 32075
            + (1461 *  (y + 4800 + n))        / 4
            + (367  *  (m - 2    - n  * 12))  / 12
            - (3    * ((y + 4900 + n) / 100)) / 4)
            < (GREGORYDAY1 - GDFIX)) {
                j += GDFIX;
            }
    }

    return(j);
}

/* convert the Julian value back to a date; be sure to pass
    writeable ints
*/
void FromJul(int j, int *y, int *m, int *d)
{
    int     i = 0;
    int     n = 0;
    int     t1 = 0;
    int     t2 = 0;
    int     ly = 0;
    int     lm = 0;

    if (j < GREGORYDAY1) {
        t1 = j + 1095;
        *y = (t1 / 1461) * 4;
        t2 =  t1 % 1461;
        ly =  t2 / 365;
        if (ly == 4) ly--;
        *d =  t2 - (ly * 365) + 1;
        *y += ly;
        *y -= 4715;
        for (i = 1; i < 12; i++) {
            if ((n = EndOfMonth(*y,i)) >= *d) break;
            *d -= n;
        }
        *m = i;
    } else {
        t1 = j + 68569;
        t2 = (4 * t1) / 146097;
        t1 = t1 - (146097 * t2 + 3) / 4;
        ly = (4000 * (t1 + 1)) / 1461001;
        t1 = t1 - (1461 * ly) / 4 + 31;
        lm = (80 * t1) / 2447;
        *d = t1 - (2447 * lm) / 80;
        t1 = lm / 11;
        *m = lm + 2 - (12 * t1);
        ly += t1 + (t2 - 49) * 100;
        *y = (ly - (ly < 1));
    }
}

/* consider all three (four?) rules for determining leap year;
    we're way out front on this one:  leap-year is skipped every hundred years,
    except for every 400th year, when it is observed - that's the 'gregorian correction';
    if we skip it every 4000th year, it refines the correction and pushes
    the horizon of seasonal drift far beyond our current ability to predict it;
    calendar mavens and some astronomers are aware of this refinement, and some version
    of it is incorporated in the russian and persian calendars, but western politics
    haven't twigged to it, and may never; it certainly doesn't matter to banking
    software, i just like having a calculation that's obsessively correct.
*/
int LeapYear(int y)
{
    if (y > GDY1)
        return(((y % 4 == 0) && ((y % 4000 != 0) && ((y % 100 != 0) || (y % 400 == 0)))) ? 1 : 0);
    else
        return((abs(y) % 4 == 0) ? 1 : 0);
}

/* convert a Julian to an 8-byte string */
char *fn_fromjul(int  j)
{
    static  char    c[12];
    int     y = 0;
    int     m = 0;
    int     d = 0;

    FromJul(j ,&y ,&m ,&d);
    sprintf(c ,"%04d%02d%02d" ,y,m,d);
    return(c);
}

/* convert a regular 8-byte yyyymmdd to Julian */
int  fn_tojul(char *t)
{
    int     i = 0;
    int     j = 10;
    int     m = 0;
    int     d = 0;
    int     y = 0;
    char    cc[15]   = "";

    if (t[0] == '-') cc[i++] = '-';
    while (i < j) { if (isdigit(t[i])) cc[i] = t[i]; else break; i++; }
    if (i == j) i = 8;
    i--;
    d = (cc[i--] & 15); d += (cc[i--] & 15) * 10;
    m = (cc[i--] & 15); m += (cc[i]   & 15) * 10;
    cc[i] = '\0';
    y = atoi(cc);
    return(ToJul(y, m, d));
}

/* return the 'old' century in this window; if we're in the last part
    of the century, then the old century is the current one; otherwise,
    the old century is the one prior.
*/
int fn_oldcentury(int y)
{
    int c;
    int r;
    struct tm  *now;
    time_t      secs;

    if (y == 0) y = 70;
    time(&secs);
    now = localtime(&secs);
    r = now->tm_year % 100;
    c = now->tm_year + 1900 - r;
    return(r < y ? c - 100 : c);
}

/* make a time-stamp, a sortable numeric string */
char *daytime(char *b)
{
    static char r[40] = "";
    struct tm  *now;
    time_t      secs;
    char        m[40] = "%04d%02d%02d_%02d_%02d:%02d:%02d";

    time(&secs);
    now = localtime(&secs);
    sprintf(r,m
        ,now->tm_year+1900
        ,now->tm_mon+1
        ,now->tm_mday
        ,(now->tm_yday+7-now->tm_wday)/7+1
        ,now->tm_hour
        ,now->tm_min
        ,now->tm_sec);
    if (b) return(strcpy(b,r));
    return(r);
}

void showhlp(int typ_help) {
    void showver();
    int i = 0;

    if (typ_help==1) {
        printf(" calends [-htjlpvsmqwya] [+n] <yyyy-mm-dd> [<yyyy-mm-dd>]\n");
        printf("   - show weekday/holiday status\n\n");
        showver();
        printf("\n");
    } else if (typ_help==2) {
        printf("calends [-sw] [+n] [yyyy-mm-dd [yyyy-mm-dd]] (default=today)\n\n");
        printf("  calends accepts an optional date, or reports on 'today'.\n");
        printf("  It uses these switches to control operations:\n\n");
        printf("    -H, -? : this help\n");
        printf("    -a : concatenate all offsets with holiday status\n");
        printf("    -b : list combined holidays for the year of 'yyyy-mm-dd'\n");
        printf("    -c : use Catholic Gregorian conversion date\n");
        printf("    -e : use English Gregorian conversion date\n");
        printf("    -f : list Federal holidays for the year of 'yyyy-mm-dd'\n");
        printf("    -j : show Julian days around 'yyyy-mm-dd'\n");
        printf("    -k : list Kalends for the month of 'yyyy-mm-dd'\n");
        printf("    -l : List w/basis all holidays for the year 'yyyy-mm-dd'\n");
        printf("    -m : show the date offset within the month\n");
        printf("    -n : list NYSE holidays for the year of 'yyyy-mm-dd'\n");
        printf("    -p : lunar phases for the year of 'yyyy-mm-dd'\n");
        printf("    -q : offset within quarter\n");
        printf("    -r : list recognized holidays for the year of 'yyyy-mm-dd'\n");
        printf("    -s : set the dcl symbol CALEND_<yyyy-mm-dd> (VMS)\n");
        printf("    -t : test loop; see boundaries from Julian 0 to today\n");
        printf("    -v : verbose, show (don't set) the value for yyyy-mm-dd\n");
        printf("    -w : offset within week (Mon=0)\n");
        printf("    -x : lunar phases, in xl format (csv)\n");
        printf("    -y : offset within year\n");
        printf("    @f : add file 'f' to the Holidays array\n\n");
        printf("  Date arithmetic:\n\n");
        printf("If 2 dates (yyyy-mm-dd) are provided, calends will report\n");
        printf("the tenor (days between, or 'holding period', to a bank).\n");
        printf("If a numeric argument is prefaced with '+', calends will\n");
        printf("report on the date of the increment (which may be negative,\n");
        printf("and makes sense but looks funny, e.g.: +-60 for 60 days prior).\n\n");
        printf("  Period offset reporting:\n\n");
        printf("calends reports the days since day 1 of the period,\n");
        printf("and days until the beginning of the next period, in the\n");
        printf("format:  _000_031 (first day of a 31-day month).\n\n");
        printf("  Holiday/weekend reporting:\n\n");
        printf("The holiday report is a 3-digit numeric, 000-777,\n");
        printf("representing yesterday, today, and tomorrow.\n");
        printf("The digits are bit-masks, where 0 indicates a normal workday,\n");
        printf("1 indicates a weekend, 2 indicates a bank/Federal holiday,\n");
        printf("and 4 indicates a customer/NYSE holiday.\n");
        printf("Output from 'w', 'm', 'q', or 'y' can be combined\n");
        printf("with 's' or 'v', which will concatenate the reports.\n\n");
        printf("In the list of holidays, 'Base Date' is the earliest date\n");
        printf("on which the holiday can occur; 'DOW' is the day-of-week,\n");
        printf("which is dictated for most holidays, 0=Monday, 6=Sunday,\n");
        printf("-1=actual except float to Monday from Sunday, -2=actual\n");
        printf("w/no exceptions, -3=lunar/religious holiday function;\n");
        printf("-4=every n years on the nth year; -5=one-time full date;\n");
        printf("'Floated' shows the calendar date on which the holiday\n");
        printf("was/will be observed.  Lunar holidays are looked up by\n");
        printf("string comparison, and have programmed functions.\n");
        printf("If the logicals/environment variables KAL_FEDERAL and/or\n");
        printf("KAL_NYSE are set to existing filenames, the contents of\n");
        printf("those files will be read in lieu of calends's own arrays.\n");
        printf("The file formats must be similar to the 'Text' field of\n");
        printf("calends's internal list (text starts after '|').\n\n");
        printf("  The holidays loaded by default are:\n\n");
        printf("Federal Reserve Bank (ours):\n");
        for (i = 0; Federal[i]; i++) printf("%s\n",Federal[i]);
        printf("\nNew York Stock Exchange (theirs):\n");
        for (i = 0; NY_SE[i];   i++) printf("%s\n",NY_SE[i]);
        printf("\n  The lunar holidays recognized by calends are:\n\n");
        for (i = 0; (lunar[i].name); i++) printf("%s\n",lunar[i].name);
        printf("\n");
        printf("The names can be shortened in a KAL_ file, but must match\n");
        printf("up to the length provided.\n\n");
        showver();
        printf("\n");
    }
    exit(0);
}

void showver()
{
    printf(" calends %s, %s, %s\n",VERSION,AUTHOR,BUILD);
}
