/* dff.c - differences analyzer, for two text files input as command line
     parameters.

    The idea is to create a useful report of changes between two text
    files.  It may be run from a batch or console, with output that can
    be directed to a (third) file.  It is not compatible with diff, nor
    is it derivative of diff or VMS' DIFFERENCES.  I wanted something
    portable, and something that employed a completely different scheme
    for analyzing and synching; the big guys don't always get it right,
    so it sometimes helps to take another look.  And I had in mind some
    scripting to deal with the output dff can generate.

    See the 'syntax' label at the end of main() for the list of switches.

    Unmatched or relocated lines in file 1 are flagged "<<<-", those in file
    2, "->>>".  To provide context in the output, one line before and one or
    more lines after all mis-matches are flagged with a colon, unless -b is
    chosen.

    Options and parameters may be typed in any order, or grouped ( -ibw, e.g.).
    The m, c, or t switches end a group, since they have trailing values
    of indeterminate length.

    Comment-filtering:  DFF accomodates either of the usual (modern) comment
    styles, a lead-in delimiter with a companion trailing delimiter, or just a
    lead-in, with the newline ending the comment, but not both styles in the
    same pass.  Cobol/Fortran style comment-column switches are not supported.
    For batches or single-language situations, the comment delimiters can be
    stored in the environment variables DFF_C and DFF_T instead of restated on
    each command-line.  The default for DFF_T is the newline terminator, so if
    you're dffing DCL or TCL, just define DFF_C.  Be careful not to leave a
    terminator value in DFF_T if you switch from C-style comments to C++
    style.  When comments are filtered, comments do not trigger a mismatch,
    but they never terminate a mismatch.  So whole-line comments may show up
    in a report if they are embedded in a mismatch block, even if comments are
    being filtered.  End-of-line comments don't cause this problem, since
    there is other text on the line to evaluate.  Whitespace checking can also
    change the appearance of things, since blanks and empty lines count in
    looking for mismatches.  I didn't bother to try to filter two or more
    comment-styles in the same pass.

    Fri Mar 13 10:18:23 1992 73 creation
    Fri Oct 01 22:55:01 1999 274 u**x port
    Mon Jan 10 15:25:58 2000 10 comment filtering
    Tue Jan 11 19:21:39 2000 11 added 'environment comments'
    Tue Jan 18 15:51:18 2000 18 two-line re-synch; no/yes switch
    Tue Feb 08 16:08:32 2000 39 settable-line re-synch; this can be
        set to 1, which will try very aggressively to match the two
        files.  Line order becomes insignificant.
    Tue Jul 11 16:32:18 2000 193 nested comments
    Wed Aug 15 13:17:02 2001 227 fixed optimization of name-only
    Sun Apr 21 19:04:00 2002 111 overhaul of dfanal; added nmatch
        trailing lines for context; switches may be grouped;
        lots of comment added to dfanal - I sort of understand how
        it works, now.
    Thu Apr 25 19:48:42 2002 115 fixed comment delimiters in quoted
        strings, also back-to-back and tightly nested com-dels:
        (* remmed-out line of code (* original comment *)*)(* CMS tag *)
    Mon May 20 09:46:40 2002 140 added strong checksum algorithm (from cksm);
    Fri Feb 10 16:14:48 2006 41 general clean-up, agreement w/cksm

    Do whatever you like with this file.  Except for the stripped-down
    standard checksum calculation, the algorithms are original, for better
    or worse.

    - Earl Roosa (earlroosa@yahoo.com)
*/

#define VERSION "2008296"
#define BUILD "Wed Oct 22 19:52:35 2008 296"
#define AUTHOR "E.Roosa"

#ifdef _MSC_VER
#include    <dos.h>
#endif
#include    <stdio.h>
#include    <stdarg.h>
#include    <stdlib.h>
#include    <string.h>
#include    <ctype.h>
#include    <sys/stat.h>

#define     ERSCAN      "scan failed"
#define     ERANAL      "comparison failed"
#define     ERTERM      "extra comment terminator at %d in:\n%s"

/* command-line switches */
#define     SWC         0x0001        /* case-insensitive match */
#define     SWD         0x0004        /* debug, show all arrays and mem ptrs */
#define     SWW         0x0002        /* space-sensitive match */
#define     SWY         0x0008        /* yes/no, just see if they match */
#define     SWN         0x0080        /* no/yes, just see if they match */
#define     SWB         0x0010        /* bare diffs */
#define     SWH         0x0020        /* help */
#define     SWF         0x0040        /* show flag array */
#define     SWS         0x0100        /* fake checksums for line values */
#define     SWQ         0x0200        /* quiet, suppress error messages */

/* print flags */
#define     PF1         0xF0        /* any flag for file 1 */
#define     PF2         0x0F        /* any flag for file 2 */
#define     PR1         0x10        /* mismatch/print line from file 1 */
#define     PR2         0x01        /* mismatch/print line from file 2 */
#define     PM1         0x20        /* match/print line from file 1 */
#define     PM2         0x02        /* match/print line from file 2 */

char        cm[6] = "";             /* comment delimiter */
char        ct[6] = "\n";           /* comment terminator */

int         nmtch = 2;              /* at least 2 lines after a mismatch */

char        f1[256] = "";           /* compare file 1 */
char        f2[256] = "";           /* compare file 2 */
char        fn[256] = "";           /* name w/o path, from f1 */
char        fp[256] = "";           /* path only, f1 */

static unsigned long const crctab[256] =
{
 0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9, 0x130476dc, 0x17c56b6b, 0x1a864db2, 0x1e475005,
 0x2608edb8, 0x22c9f00f, 0x2f8ad6d6, 0x2b4bcb61, 0x350c9b64, 0x31cd86d3, 0x3c8ea00a, 0x384fbdbd,
 0x4c11db70, 0x48d0c6c7, 0x4593e01e, 0x4152fda9, 0x5f15adac, 0x5bd4b01b, 0x569796c2, 0x52568b75,
 0x6a1936c8, 0x6ed82b7f, 0x639b0da6, 0x675a1011, 0x791d4014, 0x7ddc5da3, 0x709f7b7a, 0x745e66cd,
 0x9823b6e0, 0x9ce2ab57, 0x91a18d8e, 0x95609039, 0x8b27c03c, 0x8fe6dd8b, 0x82a5fb52, 0x8664e6e5,
 0xbe2b5b58, 0xbaea46ef, 0xb7a96036, 0xb3687d81, 0xad2f2d84, 0xa9ee3033, 0xa4ad16ea, 0xa06c0b5d,
 0xd4326d90, 0xd0f37027, 0xddb056fe, 0xd9714b49, 0xc7361b4c, 0xc3f706fb, 0xceb42022, 0xca753d95,
 0xf23a8028, 0xf6fb9d9f, 0xfbb8bb46, 0xff79a6f1, 0xe13ef6f4, 0xe5ffeb43, 0xe8bccd9a, 0xec7dd02d,
 0x34867077, 0x30476dc0, 0x3d044b19, 0x39c556ae, 0x278206ab, 0x23431b1c, 0x2e003dc5, 0x2ac12072,
 0x128e9dcf, 0x164f8078, 0x1b0ca6a1, 0x1fcdbb16, 0x018aeb13, 0x054bf6a4, 0x0808d07d, 0x0cc9cdca,
 0x7897ab07, 0x7c56b6b0, 0x71159069, 0x75d48dde, 0x6b93dddb, 0x6f52c06c, 0x6211e6b5, 0x66d0fb02,
 0x5e9f46bf, 0x5a5e5b08, 0x571d7dd1, 0x53dc6066, 0x4d9b3063, 0x495a2dd4, 0x44190b0d, 0x40d816ba,
 0xaca5c697, 0xa864db20, 0xa527fdf9, 0xa1e6e04e, 0xbfa1b04b, 0xbb60adfc, 0xb6238b25, 0xb2e29692,
 0x8aad2b2f, 0x8e6c3698, 0x832f1041, 0x87ee0df6, 0x99a95df3, 0x9d684044, 0x902b669d, 0x94ea7b2a,
 0xe0b41de7, 0xe4750050, 0xe9362689, 0xedf73b3e, 0xf3b06b3b, 0xf771768c, 0xfa325055, 0xfef34de2,
 0xc6bcf05f, 0xc27dede8, 0xcf3ecb31, 0xcbffd686, 0xd5b88683, 0xd1799b34, 0xdc3abded, 0xd8fba05a,
 0x690ce0ee, 0x6dcdfd59, 0x608edb80, 0x644fc637, 0x7a089632, 0x7ec98b85, 0x738aad5c, 0x774bb0eb,
 0x4f040d56, 0x4bc510e1, 0x46863638, 0x42472b8f, 0x5c007b8a, 0x58c1663d, 0x558240e4, 0x51435d53,
 0x251d3b9e, 0x21dc2629, 0x2c9f00f0, 0x285e1d47, 0x36194d42, 0x32d850f5, 0x3f9b762c, 0x3b5a6b9b,
 0x0315d626, 0x07d4cb91, 0x0a97ed48, 0x0e56f0ff, 0x1011a0fa, 0x14d0bd4d, 0x19939b94, 0x1d528623,
 0xf12f560e, 0xf5ee4bb9, 0xf8ad6d60, 0xfc6c70d7, 0xe22b20d2, 0xe6ea3d65, 0xeba91bbc, 0xef68060b,
 0xd727bbb6, 0xd3e6a601, 0xdea580d8, 0xda649d6f, 0xc423cd6a, 0xc0e2d0dd, 0xcda1f604, 0xc960ebb3,
 0xbd3e8d7e, 0xb9ff90c9, 0xb4bcb610, 0xb07daba7, 0xae3afba2, 0xaafbe615, 0xa7b8c0cc, 0xa379dd7b,
 0x9b3660c6, 0x9ff77d71, 0x92b45ba8, 0x9675461f, 0x8832161a, 0x8cf30bad, 0x81b02d74, 0x857130c3,
 0x5d8a9099, 0x594b8d2e, 0x5408abf7, 0x50c9b640, 0x4e8ee645, 0x4a4ffbf2, 0x470cdd2b, 0x43cdc09c,
 0x7b827d21, 0x7f436096, 0x7200464f, 0x76c15bf8, 0x68860bfd, 0x6c47164a, 0x61043093, 0x65c52d24,
 0x119b4be9, 0x155a565e, 0x18197087, 0x1cd86d30, 0x029f3d35, 0x065e2082, 0x0b1d065b, 0x0fdc1bec,
 0x3793a651, 0x3352bbe6, 0x3e119d3f, 0x3ad08088, 0x2497d08d, 0x2056cd3a, 0x2d15ebe3, 0x29d4f654,
 0xc5a92679, 0xc1683bce, 0xcc2b1d17, 0xc8ea00a0, 0xd6ad50a5, 0xd26c4d12, 0xdf2f6bcb, 0xdbee767c,
 0xe3a1cbc1, 0xe760d676, 0xea23f0af, 0xeee2ed18, 0xf0a5bd1d, 0xf464a0aa, 0xf9278673, 0xfde69bc4,
 0x89b8fd09, 0x8d79e0be, 0x803ac667, 0x84fbdbd0, 0x9abc8bd5, 0x9e7d9662, 0x933eb0bb, 0x97ffad0c,
 0xafb010b1, 0xab710d06, 0xa6322bdf, 0xa2f33668, 0xbcb4666d, 0xb8757bda, 0xb5365d03, 0xb1f740b4,
};

int main(int argc, char *argv[])
{
    long    *ck1;           /* checksum pointer block, file 1 */
    long    *ck2;           /*  "  , file 2  */
    int     i   = 0;        /* loop counter */
    int     j   = 0;        /* loop counter */
    int     lc1 = 0;        /* line count, file 1 */
    int     lc2 = 0;        /*  "  , file 2  */
    int     ln  = 0;        /* larger line count */
    int     sw  = 0;        /* switches */
    int     xtra= 0;        /* check for comment errors */
    char    c   = '\0';     /* argument switch character */
    char    *lp;            /* pointer to line status block */
    char    *env;           /* pointer to environment sym/log */
    FILE    *fo  = NULL;    /* output file pointer, for re-direct */
    char    m[2] = "w";     /* 'w'rite mode (changes to 'a'ppend) */
    char    rn[256] = "";   /* outfile name (redirect) */
    struct  stat sta;
    int     s   = 0;        /* return status */
    char    *ext1 = NULL;   /* filename extensions */
    char    *ext2 = NULL;   /* filename extensions */

    long *dfscanf(char *,int,int *,int *);
    char *dfanal(int,int,long *,long *,int);
    int dfprint(char *,char *,char *,int,int);

    /* argument check; first, see that we have two filenames;
        then load the switches (if any) into the bits of 'sw',
        to be passed to the functions.
    */
    if (argc < 2) goto syntax;

    if ((env = getenv("DFF_C")) != NULL) strcpy(cm,env);
    if ((env = getenv("DFF_T")) != NULL) strcpy(ct,env);

    for (i = 1; i < argc; i++) {
        if ((argv[i][0] == '>') || (argv[i][0] == 125)) {
            if ((argv[i][1] == '>') || (argv[i][1] == 125)) {
                strcpy(rn ,&argv[i][2]);
                m[0] = 'a';
            }
            else {
                strcpy(rn ,&argv[i][1]);
                m[0] = 'w';
            }
            if ((fo=freopen(rn ,m ,stdout)) == NULL) {
                printf("\n!! cannot redirect to %s\n" ,rn);
                exit(8);
            }
        } else if (argv[i][0] == '-') {
            for (j = 1; j < strlen(argv[i]); j++) {
                switch (c = tolower(argv[i][j])) {
                    case 'b': sw |= SWB;                        break;          /* bare, diffs only */
                    case 'c': strcpy(cm,&argv[i][++j]);     j = strlen(argv[i]); break;
                    case 'd': sw |= SWD; printf("debug on\n");  break;
                    case '?':
                    case 'h': sw |= (SWH); goto syntax;                         /* help */
                    case 'i': sw |= SWC;                        break;          /* ignore case */
                    case 'm': nmtch = atoi(&argv[i][++j]);  j = strlen(argv[i]); break;
                    case 'n': sw |= (SWN | SWB);                break;          /* no  diffs, name only */
                    case 'q': sw |= SWQ;                        break;          /* suppress errors */
                    case 's': sw |= SWS;                        break;          /* use my cheesy checksums */
                    case 't': strcpy(ct,&argv[i][++j]);     j = strlen(argv[i]); break;
                    case 'w': sw |= SWW;                        break;          /* whitespace check */
                    case 'x': sw |= SWF;                        break;          /* debug flag array only */
                    case 'y': sw |= (SWY | SWB);                break;          /* yes diffs, name only */
                    default : printf("!! %c:  unrecognized switch\n",c); break;
                }
            }
        } else if (argv[i][0] == '+') {
            if (ext1) {
                ext2 = &argv[i][1];
                strcat(f2,ext2);
            } else {
                strcpy(f2,f1);
                ext1 = &argv[i][1];
                strcat(f1,ext1);
            }
        } else if (f1[0]) {
            strncpy(f2,argv[i],255);
            if (s = stat(f2,&sta) < 0) {
                if (ispunct(f2[strlen(f2)-1])) f2[strlen(f2)-1] = '\0';
                s = stat(f2,&sta);
                strncpy(f2,argv[i],255);
            }
            if (s < 0) {
                strcpy(f2,f1);
                strncat(f2,argv[i],255-strlen(f1));
            } else if (S_ISDIR(sta.st_mode)) {
                strncpy(f2,argv[i],255-strlen(fn));
                if (!ispunct(f2[strlen(f2)-1])) strcat(f2,"/");
                strcat(f2,fn);
                if (s = stat(f2,&sta) < 0) {
                    strncpy(f2,argv[i],255-strlen(f1));
                    if (!ispunct(f2[strlen(f2)-1])) strcat(f2,"/");
                    strcat(f2,f1);
                }
            }
        } else {
            strncpy(f1,argv[i],255);
            for (s = strlen(f1); (s) && (f1[s]!=']') && (f1[s]!='\\') && (f1[s]!='/'); s--) ;
/*
                && (!ispunct(f1[s])
                || (f1[s]=='-')
                || (f1[s]=='.')
                || (f1[s]=='_')
                || (f1[s]=='$')
                ); s--) ;
*/
            strcpy(fp,f1); fp[s] = '\0'; // no use for this, yet
            if (ispunct(f1[s])) s++;
            strcpy(fn,&f1[s]);
        }
    }

    if (f2[0] <= ' ') goto syntax;

    if ((nmtch > 10) || (nmtch < 1)) nmtch = 2;
    if (sw & SWB) nmtch = 0;

    /* scan file 1, and create an array of checksums, 'ck1' */
    if ((ck1 = dfscanf(f1,sw,&lc1,&xtra)) == NULL) {
        if (!(sw & SWQ)) printf("!! %s:  ",ERSCAN);
        printf("%s\n",f1);
        exit(4);
    }
    if (xtra) printf("!! %s\n",f1);

    /* repeat, for file 2, 'ck2' */
    if ((ck2 = dfscanf(f2,sw,&lc2,&xtra)) == NULL) {
        if (!(sw & SWQ)) printf("!! %s:  ",ERSCAN);
        printf("%s\n",f2);
        exit(3);
    }
    if (xtra) printf("!! %s\n",f2);

    if (!(sw & SWB)) {
        printf("-------------------------------------------------------------------------------\n");
        printf("%30s <  > %s",f1,f2);
        printf("\n%24d lines <  > %d lines",lc1,lc2);
        if (cm[0])      printf(" .. %s comments %s",cm,ct);
        if (sw & SWC)   printf(" .. case check off");
        if (sw & SWW)   printf(" .. whitespace check");
        printf("\n");
    }

    /* scan the checksum arrays, and put the comparison results
        in a byte array, 'lp'
    */
    if ((lp = dfanal(lc1,lc2,ck1,ck2,sw)) == NULL) {
        if (sw & SWN) exit(0);
        if (sw & SWY) { printf("%s\n",f1); exit(3); }
        if (!(sw & SWQ)) {
            printf("!! %s\n",ERANAL);
            goto syntax;
        }
        exit(0);
    } else {
        if (sw & SWY) exit(0);
        if (sw & SWN) { printf("%s\n",f1); exit(0); }
    }

    ln = (lc1 > lc2 ? lc1 : lc2);
    if ((sw & SWD) || (sw & SWF))       /* if debug is on, */
        for (i = 0; i < ln; i++) {      /* print the arrays */
            printf(" %6u: ",i);
            if (i < lc1) printf(" %10lu : ",ck1[i]);
            printf(" %02X ",lp[i]);
            if (i < lc2) printf(" : %10lu ",ck2[i]);
            printf("\n");
        }

    /* print the files; 'lp' controls whether to skip or print each
        line, beginning with the first block of file 1.
    */
    exit(dfprint(f1,f2,lp,ln,sw));

syntax:

    printf("%s\n",argv[0]);
    printf("%s\n"," dff file.1 file.2 -bdinswy -m# -c<delim> -t<delim>");
    printf("%s\n"," dff -help");

    if ((sw & SWH) == 0) exit(0);

    printf("\nv.%s, %s - %s\n\n",VERSION,BUILD,AUTHOR);
    printf("%s\n"," - needs two existing files; switches are optional:");
    printf("%s\n","   -b  Bare diffs, no head/trail");
    printf("%s\n","   -c/* -t*/  Comment/Terminate");
    printf("%s\n","   -d  Debug (noisy!)");
    printf("%s\n","   -h  this Help");
    printf("%s\n","   -i  case-Insensitive");
    printf("%s\n","   -m# Match at least # lines after a mis-match");
    printf("%s\n","   -n  No/Yes, report name-only on NO difference");
    printf("%s\n","   -q  Quiet, suppress error messages");
    printf("%s\n","   -s  Speedy checksums (may report false diffs)");
    printf("%s\n","   -w  Whitespace-sensitive");
    printf("%s\n","   -y  Yes/No, report name-only on difference");
    printf("\n%s\n"," The two parameters must be the names of text files.");
    printf("%s\n"," DFF normally treats whitespace as insignificant,");
    printf("%s\n"," upper- and lower-case as different.");
    printf("%s\n"," Debug mode shows the numeric values of the arrays used");
    printf("%s\n"," in evaluating each line of text.");
    printf("%s\n"," Comments (e.g., -c/* -t*/, or \"-c!\") can be defined in");
    printf("%s\n"," the environment variables DFF_C and DFF_T, instead of on");
    printf("%s\n"," the command line.  If both are set, the command-line");
    printf("%s\n"," overrides, but be careful of comment formats without");
    printf("%s\n"," terminators; -c# and DFF_T=*/ do not work together.");

    exit(0);
}

/* function dfprint - re-reads both files, printing the lines flagged in
   the 'lp' array.
*/

int dfprint(char *f1,char *f2,char *lp,int ln,int sw)
{
    int     i1 = 0;         /* loop counters */
    int     i2 = 0;         /* loop counters */
    int     p1 = 0;         /* print counters */
    int     p2 = 0;         /* print counters */
    int     b1 = 0;         /* block counters */
    int     b2 = 0;         /* block counters */
    int     bb1 = 0;         /* block counters */
    int     bb2 = 0;         /* block counters */
    int     c  = 0;         /* character read from file */
    FILE    *f1rp;          /* file pointers */
    FILE    *f2rp;          /* file pointers */

    if ((f1rp = fopen(f1, "r")) == NULL) {
        printf("!! print open failed:  %s",f1);
        exit(2);
    }
    if ((f2rp = fopen(f2, "r")) == NULL) {
        printf("!! print open failed:  %s",f2);
        exit(2);
    }

    /* scan to the end of the 'lp' array */
    while ((i1 < ln) || (i2 < ln)) {
        /* if we haven't gone beyond the last line, and the
            lp flag says to print this side, tell the user which
            file it's from, then start printing; print until the
            flag is not on, or we get to end-of-file;
        */
        bb1 = b1;
        if ((i1 < ln) && (lp[i1] & PF1)) {
            if (!(sw & SWB)) printf("************************\n(%s)\n",f1); else printf("\n");
            b1++;
            while ((i1 < ln) && (lp[i1] & PF1)) {
                /* post-incrementing the line counter in the 'if' has the cosmetic
                    advantage of reporting the first line as '1' rather than '0'
                */
                if (lp[i1++] & PR1) {
                    p1++;                       /* count lines printed */
                    printf("%8u<<<-", i1);      /* report line numbers */
                } else {
                    printf("%8u....", i1);
                }
                while ((c = fgetc(f1rp)) != EOF) {
                    putchar(c);
                    if (c == '\n') break;
                }
            }
        }
        bb2 = b2;
        /* same operation as above, repeated on file 2 */
        if ((i2 < ln) && (lp[i2] & PF2)) {
            if (!(sw & SWB)) printf("************\n(%s)\n",f2); else printf("\n");
            b2++;
            while ((i2 < ln) && (lp[i2] & PF2)) {
                if (lp[i2++] & PR2) {
                    p2++;                       /* count lines printed */
                    printf("%8u->>>", i2);      /* report line numbers */
                } else {
                    printf("%8u....", i2);
                }
                while ((c = fgetc(f2rp)) != EOF) {
                    putchar(c);
                    if (c == '\n') break;
                }
            }
        }
        if (!(sw & SWB) && (bb1 != b1 || bb2 != b2)) {
            printf("************************\n");
        }
        /* no more printable lines in the current block of file 1;
            (or we may have started with none); start skipping;
        */
        while ((i1 < ln) && ((lp[i1] & PF1) == 0)) {
            i1++;               /* count lines skipped, also! */
            while ((c = fgetc(f1rp)) != EOF)
                if (c == '\n') break;
        }
        while ((i2 < ln) && ((lp[i2] & PF2) == 0)) {
            i2++;
            while ((c = fgetc(f2rp)) != EOF)
                if (c == '\n') break;
        }
    }
    if (!(sw & SWB)) {
        printf(" %16d lines marked < ",p1);
        printf(" > %d lines marked",p2);
        printf("\n %16d blocks       < ",b1);
        printf(" > %d blocks",b2);
        if (cm[0])      printf(" .. %s comments %s",cm,ct);
        if (sw & SWC)   printf(" .. case check off");
        if (sw & SWW)   printf(" .. whitespace check");
        printf("\n");
        printf("%30s <  > %s\n",f1,f2);
        printf("-------------------------------------------------------------------------------\n");
    }
    fclose(f1rp);
    fclose(f2rp);
    return(b1||b2);
}

/* function dfanal - scans the checksum arrays ck1 and ck2, evaluates each
   element as:  unique | exact match | first of a pair | last of a pair;
   as long as the pairs are in synch, shut off printing;
   when the match fails, back up and flag the last match as printable;
   each file has half of a byte to store its print flag, ck1 gets the
   high nibble, ck2 the low; the eval array gets passed back as a pointer
   on return.
*/

char *dfanal(int lc1,int lc2,long *ck1,long *ck2,int sw)
{
    int     m   = 0;        /* post-block matches */
    int     m1  = 0;        /* post-block matches */
    int     m2  = 0;        /* post-block matches */
    int     ln  = 0;        /* larger line count */
    int     ln1 = 0;        /* line scan number, file 1 */
    int     ln2 = 0;        /*  "  , file 2  */
    int     lx1 = 0;        /* line scan number, file 1 */
    int     lx2 = 0;        /*  "  , file 2  */
    int     y   = 0;        /* yes, they're different */
    char    *lp;

    /* allocate for the larger of the two line counts, and store the
        count for later reference
    */
    ln = (lc1 > lc2 ? lc1 : lc2);
    if ((lp = (char *)calloc(1,ln)) == NULL) {
        printf("\nlp - out of memory\n");
        return(NULL);
    }
    while ((ln1 < lc1) && (ln2 < lc2)) {
        if (sw & SWD) {                     /* if debug is on, */
            printf("%6d / ",lx1);
            printf("%6d / ",ln1);
            printf("%08x / ",ck1[ln1]);
            printf("%08x / ",ck2[ln2]);
            printf("%6d / ",ln2);
            printf("%6d \n",lx2);
        }

        if ((sw & SWW) == 0) {
            while ((ln1 < lc1) && (ck1[ln1] == 0)) ln1++;
            while ((ln2 < lc2) && (ck2[ln2] == 0)) ln2++;
        }

        /* scan the matching lines, clearing the print flags
        */
        if (lx1 || lx2) m = nmtch;
        while ((ln1 < lc1) && (ln2 < lc2) && (ck1[ln1] == ck2[ln2])) {
            /* printf("!!!! %4d: %2x .. %4d: %2x\n",ln1,lp[ln1],ln2,lp[ln2]); */
            lp[ln1] &= PF2;
            lp[ln2] &= PF1;
            if (m > 0) {
                lp[ln1] |= PM1;
                lp[ln2] |= PM2;
                m--;
            }
            ln1++;
            ln2++;
            if ((sw & SWW) == 0) {
                while ((ln1 < lc1) && (ck1[ln1] == 0)) lp[ln1++] &= PF2;
                while ((ln2 < lc2) && (ck2[ln2] == 0)) lp[ln2++] &= PF1;
            }
        }

        /* we have a mis-match, so flag the last matching line
        */
        if ((ln1 > 0) && (ln1 < lc1)) {
            if (!(sw & SWB)) lp[ln1-1] |= PM1;
        }
        if ((ln2 > 0) && (ln2 < lc2)) {
            if (!(sw & SWB)) lp[ln2-1] |= PM2;
        }

        /* start flagging mis-matches; stop if we get
            'nmatch' (two?) consecutive lines that match
        */
        m = 0;
        for (lx1 = ln1; lx1 < lc1; lx1++) {
            if (ck2[ln2+m] == ck1[lx1]) {
                if ((!nmtch) || (m++ >= nmtch)) break;
            } else if (ck1[lx1]) {
                m = 0;
                y = lp[lx1] |= PR1;
            } else {
                m = 0;
                if (!(sw & SWB)) lp[lx1] |= PM1;
            }
        }
        m = 0;
        for (lx2 = ln2; lx2 < lc2; lx2++) {
            if (ck1[ln1+m] == ck2[lx2]) {
                if ((!nmtch) || (m++ >= nmtch)) break;
            } else if (ck2[lx2]) {
                m = 0;
                y = lp[lx2] |= PR2;
            } else {
                m = 0;
                if (!(sw & SWB)) lp[lx2] |= PM2;
            }
        }
        /* printf("#### %4d: %2x .. %4d: %2x\n",lx1,lp[lx1],lx2,lp[lx2]); */

        /* if a mismatch was flagged and we just want to get a name, get out
        */
        if ((sw & (SWN|SWY)) && (y)) return(NULL);

        /* We want to keep shifting forward in the array, 'staggering' so
            that we keep trying to find matches in each file alternately.
            If 'lx' made it to the end of a file, we didn't find a match
            for the current base line; inch the base forward one line
            and try to re-synch; otherwise, jump up to where we found
            matching blocks of two (?=nmtch) lines.
        */
        if ((lx1 < lc1) && (lx2 < lc2)) {
            /* This is the one bit of magic that makes me uneasy:
                In order to re-synch after running down a match on both lines,
                we can pick either the match of the file1 line found in file2,
                or vice-versa.  How to choose?  If we set up on the line that
                occurred earlier in sequence, we could loop forever, so I'm
                using the smaller 'spread' - the match that was closer to its
                base, regardless of which base occurred earlier in the file.
            */
            if ((lx2 - ln2) < (lx1 - ln1)) {
                ln2 = lx2;
            } else {
                ln1 = lx1;
            }
        } else {
            /* This situation's easy:  we didn't match one of the lines
                we were searching for, so we just bump it up one line
                and loop around to try again.  The array from this point to
                the end has been flagged as a mismatch.  The top of the loop
                will clear it out again so we can start fresh.
            */
            if ((lx1 >= lc1) && (ln2 < lc2)) ln2++;
            if ((lx2 >= lc2) && (ln1 < lc1)) ln1++;
        }
        /* printf("@@@@ %4d: %2x .. %4d: %2x\n",ln1-1,lp[ln1-1],ln2-1,lp[ln2-1]); */
    }
    return(lp);
}

/* function dfscanf - scans input files and builds an array of 'checksum'
   totals for each line in the file; the array for each file is kept in
   core for comparison;
*/

long *dfscanf(char *f,int sw,int *l,int *xtra)
/* char    *f;                             filename */
/* int     w;                              space/case/debug */
/* int     *l;                             lines in this file (to return) */
/* int     *xtra;                          extra comment terminator? */
{
    int             c = 0;          /* character read */
    int             i = 0;          /* iterator */
    int             q = 0;          /* inside a quoted string? */
    int             s = 0;          /* size of memory block */
    unsigned long   x1 = 0;         /* checksum interim */
    unsigned long   x2 = 0;         /* checksum interim */
    unsigned long   x = 0;          /* checksum (line hash value) */
    int             y = 0;          /* file line counter */
    int             z = 0;          /* line position minder */
    FILE            *ffrp = NULL;   /* file to read */
    long            *mp = NULL;     /* memory block for checksums */
    char            b[1024] = "";   /* read buffer */
    char            *pcm = NULL;    /* pointer to comment initiator */
    char            *pct = NULL;    /* pointer to comment terminator */
    char            *pb = NULL;     /* buffer pointer */
    int             icx = 0;        /* depth of current comment context */
    int             cmn = 0;        /* comment-init length */
    int             ctn = 0;        /* comment-term length */
    int             r = 0;          /* return status */
    struct          stat stt;

    /* if debug is on, print all our business . . . */
    if (sw & SWD)
        printf("\nScanning %s\n",f);

    r = stat(f,&stt);
    if (S_ISDIR(stt.st_mode)) exit(0);

    if ((ffrp = fopen(f, "r")) == NULL) return(NULL);
#ifndef __MSDOS__
    if ((mp = (long *)malloc(1040)) == NULL) return(NULL);
#endif
    if (cm[0]) {
        cmn = strlen(cm);
        ctn = strlen(ct);
    }
    while (fgets(b,1022,ffrp)) {
        /* getc(stdin); */
        for (q = i = 0; i < 1024; i++) {
            /*
              cksm srch and dff share this algorithm:
                comments may be c-style:  intro->terminator,
                    or c++ style:  intro->endofline;
                terminator-style comments may be nested to any reasonable depth, but the
                    delimiters must balance; we yell if they don't;
                    we don't care whether the compiler permits nesting;
                delimiters included in quoted strings within code will not be processed
                    as comment delimiters - be sure to clear q at eol;
            */
            if (cmn) {
                if ((!icx) && (b[i] == 34)) q = (1 & !q);
                if (!q) while (!strncmp(cm,&b[i],cmn) || !strncmp(ct,&b[i],ctn)) {
                    if (!strncmp(cm,&b[i],cmn)) {
                        icx++;
                        i += cmn;
                    } else if (!strncmp(ct,&b[i],ctn)) {
                        if (icx > 0) {
                            icx--;
                        } else if ((ct[0] != '\n') && !(sw & SWQ)) {
                            printf(ERTERM,i,b);
                            *xtra = 1;
                        }
                        if (ct[0] == '\n') break; else i += ctn;
                    }
                }
            }
            c = b[i] & 0xff;
            if (c < 1) break;
            /* skip whitespace unless ordered not to:
            */
            if ((icx == 0) && ((sw & SWW)
            || ((c != '\n') && (c != '\t') && (c != ' ') && (c != '\r')))) {
                /* low-case the uppers if told to */
                if ((sw & SWC) && (c <= 'Z') && (c >= 'A')) c |= 32;
                if (sw & SWS) {
                    x += (c << (z & 7));
                } else {
                    x1 = (x >> 24) ^ c;
                    x2 = x << 8;
                    x = x2 ^ crctab[x1 & 0xFF];
                }
                z++;
            }
            /* at every line end, first check to see if we need to
              allocate some core (first time, for sure we do); grab 1k
              at a time, or 256 lines worth; tell the debugging user all
              about it, since he asked;
            */
            if (c == '\n') {
                if ((y & 255) == 0) {
                    s = (y+256) * sizeof(x) + 16;
                    if (sw & SWD) {
                        printf("\rallocating %d bytes,",s);
                        printf(" lines read:  %d",y);
                    }
                    if ((mp = (long *)realloc(mp,s)) == NULL) {
                        printf("\nmp - out of memory\n");
                        return(NULL);
                    }
                    if (sw & SWD) printf(" mp = %p\n",mp);
                }
                /* printf(".%d = %d.",y,x); */
                if (sw & SWD) printf("#%6d (%08x) %s",z,x,b);
                mp[y++] = x;
                z = 0;
                x = 0;
                if (ct[0] == '\n') icx = 0;
                /* printf("%c %d %d...",c,z,x); */
                break;
            }
        }
    }
    mp[y] = 0;                  /* terminate the array */
    *l = y;                     /* pass back the line count, */
    fclose(ffrp);
    return(mp);                 /*   and the array pointer. */
}
