/* hlp2html.c - convert a vms help file to a standalone page of html

    vms help files have a very simple text format:

1 main_topic

    Some text, plain ascii.  Wrap lines wherever you like, html won't
    care where you put the line-end.  Each line of text has to be
    indented at least 1 space, only level numbers go on the margin.

    Continue putting in more paragraphs, until you get to the next
    topic or subtopic.

2 sub-topic_1

    this fits under main_topic, and will be listed in the links under
    main_topc.

3 sub-sub-topic

    etc.

2 sub-topic_2

    you get the idea.  nest as many levels as you like (i think the vms
    help level is 16 or so, but nobody's gonna read a help file that deep).
    the output isn't fancy, but the links work, and there's a return link
    at each sub to go back to top-of-page.

Additional rules:

  . anything indented more than 4 spaces is formatted in monospace, assuming a list or table:
    if you indent to 4 or less, hlp2html will just output default text;
    indent 5 or more, and the text will be monospaced, which i find useful
    for presenting stack-dumps, tables, program listings, etc.

  . non-numbered list:
    *
    - bullet 1
    - bullet 2
    .*

  . numbered list:
    #
    - item 1
    - item 2
    .#

    this is free source; do whatever you like with it . . .

    Mon Jan 16 15:33:44 2006 16 - earlroosa@yahoo.com
*/

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#ifdef VMS
#include <types.h>
#include <ctype.h>
#include <stat.h>
#else
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

main(int argc, char *argv[])
{
    int     i = 0;
    FILE    *rf = NULL;
    FILE    *wf = NULL;
    char    rn[256] = "";
    char    wn[256] = "";
    void readwrite(FILE *rf, FILE *wf);

    if (argc < 3) {
        printf("\n hlp2html <vmshelpfile> <htmlfile> [title_text]\n\n");
        printf("    format:\n");
        printf("    1 At_the_margin:__topic_(no_spaces)\n");
        printf("        indent up to 4 spaces:  default text\n");
        printf("            indent further:  monospace, for tables, etc.\n");
        printf("        #\n");
        printf("        - item in a numbered list\n");
        printf("        .#\n");
        printf("        *\n");
        printf("        - item in a bulleted list\n");
        printf("        .*\n");
        printf("    2 Nested_topic_(depth_to_15)\n");
        exit(0);
    }
    if ((rf = fopen(argv[1],"r")) == NULL) {
        printf("!! can't open %s\n",argv[1]);
        exit(3);
    }
    if ((wf = fopen(argv[2],"w")) == NULL) {
        printf("!! can't open %s\n",argv[2]);
        exit(3);
    }
    if (argc > 3) {
        fprintf(wf,"<html><head><title>%s</title></head><body>\n",argv[3]);
    }
    readwrite(rf,wf);
    if (argc > 3) {
        fprintf(wf,"</body></html>\n");
    }
    fclose(rf);
    fclose(wf);
    exit(0);
}

void readwrite(FILE *rf, FILE *wf)
{
    int     i = 0;      /* iterator */
    int     j = 0;      /* iterator */
    int     c = 0;      /* close paragraph w/pre? */
    int     s = 0;      /* paragraph started? */
    int     f = 0;      /* link-list started? */
    int     p = 0;      /* file position */
    char    l = 0;      /* new level */
    char    ll = '1';   /* old level */
    char    b[1024] = "";
    char    h[1024] = "";
    char    *pp[] = {"<p>","<pre>"};
    char    *cp[] = {"</p>","</pre>"};
    char    *li[] = {"</ul>","</ol>"};

    char *escape(char *b);
    char *format(char *b);
    char *hlink(char *b);

    while (fgets(b,1023,rf)) {
        if (b[0] == '\n') {
            fprintf(wf,"%s\n",cp[c]);
            s = 0;
        } else if (b[0] == ' ') {
            if (!s) {
                for (j = 1; j < 6; j++) if (b[j] > ' ') break;
                c = (j > 4);
                fprintf(wf,"%s",pp[c]);
                s = 1;
            }
            switch (b[j]) {
                case '<':   fprintf(wf,"%s\n",b); break;
                case '*':   fprintf(wf,"<ul>\n"); break;
                case '#':   fprintf(wf,"<ol>\n"); break;
                case '.':   fprintf(wf,"%s\n",li[(b[j+1]=='#')]); break;
                case '-':   fprintf(wf,"<li>%s\n",&b[j+2]); break;
                default:    fprintf(wf,"%s\n",escape(b));
            }
        } else {
            if (isdigit(b[0])) {
                l = b[0];
            } else {
                printf("!! bad help file format:\n%s\n",b);
                exit(3);
            }
            if (l > ll) {
                f = 0;
                do {
                    if (b[0] <= ' ') ;
                    else if (b[0] < l) break;
                    else if (b[0] == l) {
                        if (!f) {
                            fprintf(wf,"<p><font size=3>Additional information:</font></p>\n%s\n",hlink(escape(b)));
                            f = 1;
                        } else {
                            fprintf(wf,"&nbsp&nbsp&nbsp%s\n",hlink(escape(b)));
                        }
                    }
                } while (fgets(b,1023,rf));
                if (fseek(rf,p,0)) {
                    printf("!! seek failed at %d:\n%s\n",p,b);
                    exit(3);
                }
                fgets(b,1023,rf);
            } else {
                if (l == '1') strcpy(h,hlink(escape(b)));
            }
            if (ll > '1')
                fprintf(wf,"\n\n<p><center><font size=3>  Return to Top:  %s</font></center></p>\n\n<hr>",h);
            else if (l > '1')
                fprintf(wf,"\n<p></p><hr>");
            ll = l;
            fprintf(wf,"%s\n",format(escape(b)));
            s = 0;
        }
        p = ftell(rf);
    }
    if (l != 0) fprintf(wf,"\n\n<p><center><font size=3>  Return to Top:  %s</font></center></p>\n\n<hr>\n",h);
}

char *escape(char *b)
{
    int     i = 0;
    int     j = 0;
    int     k = 0;
    char    *e = "<>";
    char    *h[] = {"&lt;","&gt;"};
    static  char r[1024] = "";

    for (i = 0; i < strlen(b); i++) {
        if ((b[i] >= ' ') || (b[i] == '\t')) {
            for (j = 0; j < strlen(e); j++) {
                if (b[i] == e[j]) {
                    strcat(r,h[j]);
                    k = strlen(r);
                    break;
                }
            }
            if (!e[j]) {
                r[k++] = b[i];
                r[k] = '\0';
            }
        }
    }
    return(r);
}

char *hlink(char *b)
{
    int     i = 0;
    int     j = 0;
    int     s = 0;
    static  char r[1024] = "";

    strcpy(r,"<tt><strong><a href=\"#");
    j = strlen(r);
    for (i = 0; i < strlen(b); i++) {
        if (b[i] == ' ') {
            r[j] = '_';
            s = -1;
        } else {
            r[j] = b[i];
            if (s < 0) s = i;
        }
        j++;
    }
    r[j++] = '\"';
    r[j] = '\0';
    strcat(r,">");
    strcat(r,&b[s]);
    strcat(r,"</a></font>    </strong></tt>");
    return(r);
}

char *format(char *b)
{
    int     i = 0;
    int     j = 0;
    int     s = 0;
    static  char r[1024] = "";

    strcpy(r,"<p><font size=4><a name=\"");
    j = strlen(r);
    for (i = 0; i < strlen(b); i++) {
        if (b[i] == ' ') {
            r[j] = '_';
            s = -1;
        } else {
            r[j] = b[i];
            if (s < 0) s = i;
        }
        j++;
    }
    r[j++] = '\"';
    r[j] = '\0';
    strcat(r,">");
    strcat(r,&b[s]);
    strcat(r,"</a></font>");
    return(r);
}
