/*****************************************************************
 srch.c - search for some strings in some files (see the help message)

    Usage:

        srch -abCgimq0 -c(*_*) ,str1 ,str2 . . . file1.mod file2.def . . .

    See the function srx_help at the bottom of this file for online help.

    Features:

        -   online help
        -   wildcard file processing on NT, *U*X, VMS
        -   unlimited search-list (OR list)
        -   multi-string matches (AND list)
        -   exclusion strings to limit false hits
        -   search-and-replace
        -   search binary files for embedded strings
        -   user-controlled context window, 0 - 100 lines
        -   strip or include comments in searches
        -   comment exclusion by default for:  .c,.cxx,.java,.tcl,.bas,.mod
        -   user-specifiable comment delimiters for unknown file types
        -   selectable aggressive search modes
        -   finite-state-machine to optimize multi-string searches
        -   search-in-position

    Mon Feb 10 14:26:15 1997 41 - release 0, e.roosa
    Thu Feb 27 20:11:45 1997 58 - 1.1, statement-cat for .c, comments, .for
    Tue Mar 11 15:52:40 1997 70 - 1.1a, flg bit-map, for option reset
    Wed Mar 12 19:31:01 1997 71 - 1.1b, bug-fix for line overruns (accvio)
    Sat Mar 15 08:54:28 1997 74 - 1.1c, longest-line report
    Sun Mar 16 21:11:10 1997 75 - 1.2, console-entry words, dos dir fix
    Wed Apr 30 11:20:03 1997 120- 1.3, bug-fix for multi-find, w/excludes
    Tue Dec 30 22:16:11 1997 364- 2.0, added '+' lists
    Thu Aug 13 12:12:10 1998 225- 3.0, string replacements
    3.1:  switches -x (executable), -m (machine/binary) -g (aggressive)
    3.2:  functions srch_lst, srch_add, srch_help split out
    3.2a,b:  bug fixes for findfile, comment processing
    Mon Mar 18 22:23:48 2002 77 - 3.2c, -f filenames w/bare text
    v4 started date-based version numbers
    Sun Apr 28 18:18:41 2002 118
    Tue Apr 30 21:08:15 2002 120 v4 ready
    Wed May  1 09:47:23 2002 121 removed -x, rearranged the arg loop
    Thu May  2 15:21:26 2002 122 moved vms find_file up to main; -w0, -q fixed
    Thu May 16 21:23:09 2002 136 cleanup of srx_hit for multi-hit w/exclusions
    Sat May 18 12:33:41 2002 138 state machine added, srx_match/srx_mat
    Sun Jun  2 11:39:44 2002 153 search-in-position
    Mon Jun  3 19:21:06 2002 154 multi-function srx_scan
    Wed Jul 11 15:22:00 2007 192 wide-character non-support (we ignore the nulls);
        this can be very useful in searching silly not-really-utf8 files like the
        output of regedt;
        'quiet' mode finally works:  filenames, with hits, but no header/trailer;
        'dr' is a new companion to srch:  it walks down the directory tree from the
        current default, and executes the given command using the current file in
        the tree as a parameter - almost a '-r'ecursive switch; see dr.c;

    This program is free software, do whatever you like with it.

    I will try to respond to emails with questions or comments - I like mail,
    even flames, if you're so inclined.

    There is
    absolutely no warranty or assurance of suitability of this code for any
    purpose whatsoever.

*/

#define VERSION "2008235"
#define BUILD "Fri Aug 22 14:21:18 2008 235"
#define AUTHOR "EarlRoosa@yahoo.com"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <dirent.h>
//#include <fnmatch.h>

#ifdef _MSC_VER
#include <malloc.h>
#include <dos.h>
#ifdef _WIN32
#include <io.h>
#include <direct.h>
#else
#include <dir.h>
#endif
#endif

#ifdef VMS
#include <descrip.h>
#include <types.h>
#include <stat.h>
#include <unixio.h>
#else
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif

#define TRUE    1
#define FALSE   0

#define MAXLIN 32000

#define _A_  0x0001 /* append before scan */
#define _B_  0x0002 /* bare, hits only */
#define _C_  0x0004 /* include comments */
#define _F_  0x0010 /* filenames only */
#define _G_  0x0028 /* aggressive, no w/s, downcase, no punct */
#define _H_  0x0100 /* header, not */
#define _I_  0x0008 /* ignore case */
#define _M_  0x0200 /* machine code */
#define _N_  0x2000 /* no-hit filenames only */
#define _P_  0x1000 /* find only @ position */
#define _Q_  0x0012 /* quiet, hits+names */
#define _R_  0x0186 /* replace, w/ */
#define _S_  0x0080 /*  search */
#define _T_  0x0400 /* turtle mode, no state machine */
#define _U_  0x4000 /* unsigned (high-bit) chrs ok */
#define _W_  0x0800 /* set window-size */
#define _X_  0x8000 /* utf-16, wide chars, skip nulls */
#define _Z_  0x0040 /* zero-hit files reported */

    time_t  tim;

    char    words[256]  = "stdin";      /* stdin, or word-list file */
    char    **xpa       = NULL;         /* exclusion list */
    int     xl          = 0;            /* list length */
    char    **zpa       = NULL;         /* peremptory exclusions */
    int     zl          = 0;            /* list length */
    char    **vpa       = NULL;         /* search list */
    int     vl          = 0;            /* list length */
    char    **rpa       = NULL;         /* replacement list */
    int     rl          = 0;            /* list length */

    int     **fsm       = NULL;         /* state machine:  see srx_match */
    int     **fsx       = NULL;         /* exclusion state machine */

    int     fsz         = 0;            /* state machine size */
    int     fst         = 0;            /* current state */
    int     *fsh        = NULL;         /* hit list */

    int     invert      = 0;            /* invert the sense of a 'hit' */

    char            filename[256]   = "stdin";      /* current (last) filename */
    unsigned long   files           = 0;            /* global stats */
    unsigned long   lines           = 0;
    unsigned long   longest         = 0;
    unsigned long   scanned         = 0;
    unsigned long   hitcount        = 0;

    char    hmsk[32]    = "%7d %-15s-%c";   /* the hit-mask (re-written by srx_words) */
    char    pmsk[32]    = "%7d %-15s  ";    /* post-hit context lines */

    int     psn         = -1;           /* search at position n only */
    int     psx         = -1;           /* exclude at position n only */

    char    ftag[1024]  = "";           /* an optional 'tag' to find in filenames */
    char    ntag[1024]  = "";           /* an optional 'tag' to NOT find in filenames */

int main(int argc, char *argv[])
{
    int     i           = 0;            /* iterators */
    int     j           = 0;            /*     "     */
    int     k           = 0;            /*     "     */
    char    c           = '\0';         /* argument switch character */
    int     sw          = 0;            /* switches */
    int     dbg         = 0;            /* debug level */
    int     lln         = 0;            /* binary line-length */
    int     wnd         = 1;            /* window size (context lines) */
    int     fct         = 0;            /* file count (files read) */
    int     endargs     = 0;            /* end of switches and commandline search-words */
    char    fn[256]     = "";           /* output filename */
    char    fs[256]     = "";           /* filespec to search */
    FILE    *fp         = NULL;         /* output file pointer, for re-direct */
    char    m[2]        = "w";          /* 'w'rite mode (changes to 'a'ppend) */
    char    cmnt[10]    = "";           /* command-line comment */
    char    nwr[24]     = "";           /* since this time (mod-time Before or After) */
    time_t  nsec        = 0;
    int     old         = 0;
    struct  tm ttim;
    char    *env        = NULL;         /* pointer to environment symbol */
    int     recr        = 0;            /* recursive directory search */
#ifdef VMS
    char    fname[256]  = "";           /* returned from  lib$find_file */
    int     cx          = 0;            /* 'context', for lib$find_file */
    $DESCRIPTOR(fret,fname);
    $DESCRIPTOR(finp,fs);
    extern int LIB$FIND_FILE();
#else
    void    fnm_split (char *fs, char *dr, char *nm);
    int     fnm_match (char *pattern, char *fs, int flags);
#endif
    void    srx_help(int sw);
    int     srx_read(char *fspec, char *fn, int fct, char *cmnt, int sw, int wnd, int lln, int dbg, int old, time_t nsec);
    int     srx_words(char *fn, int sw);
    int     srx_word(char *, int sw);

    if (argc < 2) {
        //printf("%s\n",argv[0]);
        srx_help(0);
    }
    for (i = 1; i < argc; i++) {
        //printf("argc %d %d %s\n",i,argc,argv[i]);
        if ((argv[i][0] == '>') || (argv[i][0] == 125)) {
            if ((argv[i][1] == '>') || (argv[i][1] == 125)) {
                strcpy(fn,&argv[i][2]);
                m[0] = 'a';
            }
            else {
                strcpy(fn,&argv[i][1]);
                m[0] = 'w';
            }
            endargs = 1;
        } else if ((argv[i][0] == '<') || (argv[i][0] == '@') || (argv[i][0] == 123)) {
            sw |= _T_ & srx_words(&argv[i][1],sw);                              /* search list */
            endargs = 1;
        } else if (argv[i][0] == '-') {
            for (j = 1; j < strlen(argv[i]); j++) {
                switch (c = argv[i][j]) {
                    case '-': endargs = 1;                      break;
                    case '!': invert = 1;                       break;
                    case '?':
                    case 'h': srx_help(1);                                      /* Help */
                    case 'H': srx_help(2);                                      /* more Help */
                    case 'a': sw |= _A_;                        break;          /* Append extended lines of C */
                    case 'b': sw |= (_B_ | _H_);                break;          /* Bare, hits only */
                    case 'B': old = -1;
                    case 'A':
                              strcpy(nwr,&argv[i][j+1]);
                              ttim.tm_sec  = atoi(&nwr[18]);    nwr[17] = '\0';
                              ttim.tm_min  = atoi(&nwr[15]);    nwr[14] = '\0';
                              ttim.tm_hour = atoi(&nwr[12]);    nwr[8]  = '\0';
                              ttim.tm_mday = atoi(&nwr[6]);     nwr[6]  = '\0';
                              ttim.tm_mon  = atoi(&nwr[4]) - 1; nwr[4]  = '\0';
                              ttim.tm_year = atoi(nwr) - 1900;
                              nsec = mktime(&ttim);
                              j = strlen(argv[i]);
                              break;
                    case 'c': sw |= _C_;                        break;          /* include Comments */
                    case 'y':
                    case 'f': sw |= (_F_ | _H_); wnd = 0;       break;          /* Filenames only */
                    case 'g': sw |= _G_;                        break;          /* aGgressive search */
                    case 'i': sw |= _I_;                        break;          /* Ignore case */
                    case 'm': sw |= _M_;                        break;          /* binary (Machine code) */
                    case 'n': sw |= (_N_ | _Z_ | _H_); wnd = 0; break;          /* report Zero-hit filenames only */
                    case 'q': sw |= (_F_ | _H_ | _Q_);          break;          /* Quiet, shut off header/trailer */
                    case 'r': recr = 1;                         break;          /* recurse into directories */
                    case 't': sw |= _T_;                        break;          /* Turtle mode, don't use fsm */
                    case 'u':
                    case '8': sw |= _U_;                        break;          /* unsigned-char (8-bit) ok */
                    case 'X': sw |= _X_;                        break;          /* wide chars, skip nulls */
                    case '0':
                    case 'z': sw |= _Z_;                        break;          /* report Zero-hit files (plus normal hits) */
                    case 'C': strcpy(cmnt,&argv[i][++j]);   j = strlen(argv[i]); break; /* Comment delims */
                    case 'F': strcpy(fs,&argv[i][++j]);     j = strlen(argv[i]); break; /* File to search */
                    case 'O': strcpy(fn,&argv[i][++j]);     j = strlen(argv[i]); break; /* Output file */
                    case '*': strcpy(ftag,&argv[i][++j]);   j = strlen(argv[i]); break; /* filenam tag */
                    case '#': strcpy(ntag,&argv[i][++j]);   j = strlen(argv[i]); break; /* filenam tag, NOT */
                    case '@':
                    case 'S': sw |= _T_ &
                              srx_words(&argv[i][++j],sw);  j = strlen(argv[i]); break; /* Search list */
                    case 'd': dbg = atoi(&argv[i][++j]);    j = strlen(argv[i]); break; /* Debug level */
                    case 'l': lln = atoi(&argv[i][++j]);    j = strlen(argv[i]); break; /* bin line Length */
                    case 'w': wnd = atoi(&argv[i][++j]);    j = strlen(argv[i]); break; /* context Window size */
                    case 'p': sw |= (_P_|_T_);
                              psn = atoi(&argv[i][++j]);    j = strlen(argv[i]); break; /* search only at Position # */
                    case 'x': sw |= (_P_|_T_);
                              psx = atoi(&argv[i][++j]);    j = strlen(argv[i]); break; /* eXclude only at position # */
                    default : printf("!! %c:  unrecognized switch\n",c);         break;
                }
                if (wnd < 1) { sw |= (_W_|_Q_); wnd = 1; }
            }
        } else if (argv[i][0] == ',') {
            sw |= srx_word(&argv[i][1],sw);
            endargs = 1;
        } else if (argv[i][0] == '?') {
            sw |= srx_word(argv[i],sw);
            endargs = 1;
        } else if (!endargs) {
            sw |= srx_word(argv[i],sw);
        } else {
            if (strlen(argv[i]) < 255) {
                strcpy(fs,argv[i]);
            } else {
                printf("!! argument too long:\n%s\n",argv[i]);
            }
            //printf("+--->%d %s %s\n",i,argv[i],fs);
        }
        //printf("---->%d %s %s\n",i,argv[i],fs);
#ifndef VMS
        if (recr) {
            fnm_split(fs,fs,ftag);
        }
#endif
        //printf("====>%d %s..%s\n",k,fs,ftag);
        if (fn[0]) {
            if ((fp=freopen(fn ,m ,stdout)) == NULL) {
                printf("\n!! cannot redirect to %s\n" ,fn);
                exit(8);
            }
            fn[0] = '\0';
        }
        if (fs[0]) {
            strcpy(filename,fs);
            if (!vpa) sw |= srx_words("stdin",sw);
            if (rl != vl) sw &= ~_S_;     /* make sure of the switch combination for search/replace */
            if (xl) sw |= _T_;          /* turtle mode, we have exclusion wrapper(s) */
            if (vpa) {
#ifdef VMS
                while (1 & LIB$FIND_FILE(&finp,&fret,&cx,0,0,0,0)) {
                    for (j = strlen(fname); fname[j] <= ' '; fname[j--] = '\0') ;
                    fct += srx_read(fs,fname,fct,cmnt,sw,wnd,lln,dbg,old,nsec);
                }
#else
                fct += srx_read(fs,fs,fct,cmnt,sw,wnd,lln,dbg,old,nsec);
#endif
            }
            //if (!files) {
            //    printf("!! could not read %s\n",fs);
            //    exit(errno);
            //}
        }
    }
    if (!fs[0]) {
        strcpy(fs,"stdin");
        strcpy(filename,fs);
        fct += srx_read(fs,fs,fct,cmnt,sw,wnd,lln,dbg,old,nsec);
    }
    if (!(sw & _H_))
    {
        printf("=****** %s %lu files, %lu lines (longest:  %lu), %lu scanned, %lu hits.\n"
            ,filename,files,lines,longest,scanned,hitcount);
        time(&tim);
        printf("------- %s\n",ctime(&tim));
    }
    if (fp) fclose(fp);
    exit(hitcount>0);
}

#ifndef VMS
int fnm_match (char *pattern, char *fs, int flags) {
    static char dr[256] = "";
    static char nm[256] = "";
    void fnm_split (char *fs, char *dr, char *nm);
    int wc_match(char *pTameText, char *pWildText, int bCaseSensitive, char cAltTerminator);

    if (strlen(fs) > 255) return(0);
    fnm_split(fs, dr, nm);
    return(wc_match(nm,pattern,0,'\0'));
}

void fnm_split (char *fs, char *dr, char *nm) {
    int k = 0;

    if (fs != dr) strcpy(dr,fs);
    for (k = strlen(dr); k > 0; k--) {
        if ((dr[k] == '\\')
        ||  (dr[k] == '/')) {
            dr[k++] = '\0';
            strcpy(nm,&dr[k]);
            break;
        } else {
        if ((dr[k] == ':')
        ||  (dr[k] == ']')) {
            k++;
            strcpy(nm,&dr[k]);
            dr[k] = '\0';
            break;
        }}
    }
    if (k==0) {
        strcpy(nm,dr);
//        strcpy(dr,".");
    }
}

// Matching Wildcards: An Algorithm
// by Kirk J. Krauss
//      http://www.ddj.com
// This function compares text strings, one of which can have wildcards ('*').
//
int wc_match(char *pTameText, char *pWildText, int bCaseSensitive, char cAltTerminator) {
                       // A string without wildcards
                                        // A (potentially) corresponding string with wildcards
                                                         // By default, match on 'X' vs 'x'
                                                                             // For function names, for example, 
                                                                             // you can stop at the first '('

        int  bMatch = TRUE;
        char *pAfterLastWild = NULL;  // The location after the last '*', if we've encountered one
        char *pAfterLastTame = NULL;  // The location in the tame string, from which we started after last wildcard
        char t, w;

        // Walk the text strings one character at a time.
        while (1)
        {
               t = *pTameText;
               w = *pWildText;

               // How do you match a unique text string?
               if (!t || t == cAltTerminator)
               {
                       // Easy: unique up on it!
                       if (!w || w == cAltTerminator)
                       {
                              break;                                 // "x" matches "x"
                       }
                       else if (w == '*')
                       {
                              pWildText++;
                              continue;                          // "x*" matches "x" or "xy"
                       }
                       else if (pAfterLastTame)
                       {
                              if (!(*pAfterLastTame) || *pAfterLastTame == cAltTerminator)
                              {
                                      bMatch = FALSE;
                                      break;
                              }
                              pTameText = pAfterLastTame++;
                              pWildText = pAfterLastWild;
                              continue;
                       }

                       bMatch = FALSE;
                       break;                                         // "x" doesn't match "xy"
               }
               else
               {
                       if (!bCaseSensitive)
                       {
                              // Lowercase the characters to be compared.
                              if (t >= 'A' && t <= 'Z')
                              {
                                      t += ('a' - 'A');
                              }

                              if (w >= 'A' && w <= 'Z')
                              {
                                      w += ('a' - 'A');
                              }
                       }

                       // How do you match a tame text string?
                       if (t != w)
                       {
                              // The tame way: unique up on it!
                              if (w == '*')
                              {
                                      pAfterLastWild = ++pWildText;
                                      pAfterLastTame = pTameText;
                                      continue;                          // "*y" matches "xy"
                              }
                              else if (pAfterLastWild)
                              {
                                      pWildText = pAfterLastWild;
                                      w = *pWildText;

                                      if (!w || w == cAltTerminator)
                                      {
                                             break;                          // "*" matches "x"
                                      }
                                      else
                                      {
                                             if (!bCaseSensitive && w >= 'A' && w <= 'Z')
                                             {
                                                     w += ('a' - 'A');
                                             }
                                             if (t == w)
                                             {
                                                     pWildText++;
                                             }
                                      }
                                      pTameText++;
                                      continue;                          // "*sip*" matches "mississippi"
                              }
                              else
                              {
                                      bMatch = FALSE;
                                      break;                                 // "x" doesn't match "y"
                              }
                       }
               }

               pTameText++;
               pWildText++;
        }

        return bMatch;
}
#endif

/*****************************************************************
 srx_read:  scan an input file, looking for matches on any word in the list;
*/
int srx_read(char *fspec, char *fn, int fct, char *cmnt, int sw, int wnd, int lln, int dbg, int old, time_t nsec)
{
    int     i           = 0;            /* iterators */
    int     j           = 0;            /*     "     */
    int     k           = 0;            /*     "     */
    int     s           = 0;            /* status */
    int     q           = 0;            /* inside a quoted string? */
    int     hdr         = 0;            /* file header due? */
    int     eol         = 0;            /* end-of-line, or append next? */
    char    c           = '\0';         /* input character */
    char    fname[256]  = "stdin";      /* stdin, or search file */
    char    ftime[28]   = "";           /* file creation time, from stat */
    FILE    *fp         = NULL;         /* file pointer */
    char    m[4]        = "r";          /* 'r'ead mode (changes to "rb"inary) */
    char    *vp         = NULL;         /* pointer to the search token */
    char    *b          = NULL;         /* input buffer pointer */
    int     szb         = 0;            /* strlen of b, after fgets */
    char    **bb        = NULL;         /* input buffer array */
    int     *hits       = NULL;         /* lines shown in the current window */
    int     hn          = 0;            /* hit count */
    int     bn          = 0;            /* current buffer line number */
    int     ln          = 0;            /* current file line number */
    int     mn          = 0;            /* line length */
    int     mln         = 0;            /* maximum line length */
    char    rb[MAXLIN]  = "";           /* read buffer (after append & clean) */
    int     rp          = 0;            /* read buffer position */
    int     rn          = 0;            /* number of lines read */
    int     ps          = 0;            /* postscript context lines */
    char    cm[20]      = "";           /* comment delimiter */
    char    ct[20]      = "";           /* comment terminator */
    int     cmn         = 0;            /* comment-init length */
    int     ctn         = 0;            /* comment-term length */
    int     icx         = 0;            /* depth of current comment context */
    struct  stat        sta;
    DIR     *dr;
    struct  dirent      *dn;
    char    ds[512] =    "";
    char *(*srx_scan)();

    void srx_comment(char *fn, char *comnt, char *cm, char *ct);
    char *srx_scanr(char *rb);  /* search/replace */
    char *srx_scanp(char *rb);  /* positional */
    char *srx_scant(char *rb);  /* turtle mode (strstr) */
    char *srx_scanf(char *rb);  /* fast, fsm */
    char *srx_fgets(char *b, int lln, FILE *fp, int sw);
    int srx_replace(char *rb);
    void srx_mat(int ***, int, char *);
#ifndef VMS
    int fnm_match (char *pattern, char *fs, int flags);
#endif

    /* sanity check:  _Replace and _Machine are fussy, since they call
        different scan functions and produce special-purpose output;
        replace wants to produce bare output, comments included, single lines
        of any length, so that swapping strings has no side effects;
        machine code doesn't look pretty onscreen, and we don't want to
        search high-bit strings anyway, so we're trimming out all the junk
        and printing the compressed read-buffer;
        we're doing the checks here because we might want to fiddle with
        something before we do the file open, and because it's cheap and
        visible to get it done early;
    */
    if (sw & _S_) {
        srx_scan = srx_scanr;
        sw = _R_;                               /* _R_ = _S_|_B_|_C_ */
        lln = 0;
        wnd = 1;
    } else {
        if      (sw & _P_) srx_scan = srx_scanp;
        else if (sw & _T_) srx_scan = srx_scant;
        else               srx_scan = srx_scanf;
        if (sw & _M_) {
            sw |= _C_;
            if (!lln) lln = 110;
            wnd = 1;
        }
    }
    /* sanity check II:  just in case the line length or context
        window got loaded with unreasonable values
    */
    if (!lln) lln = (MAXLIN - 2);
    if (wnd < 1) { sw |= (_W_|_Q_); wnd = 1; }
    if (wnd > 100) wnd = 2;
    /*
       III:  if we asked for a position-exclusion, but only loaded the search list
    */
    if ((sw & _P_) && (psn < 0) && (psx >= 0) && (zpa == NULL)) {
        zpa = vpa;
        zl = vl;
        vpa = NULL;
        vl = 0;
    }

    if (fn[0]) strcpy(fname,fn);
    if (dbg > 1) { fprintf(stderr,"%s\n",fspec); }

    if (!(files)) {                             /* first-file only */
        if (!rl) {
            if (!(sw & _T_)) {
                assert((fsh = (int *)calloc(vl,sizeof(int))) != NULL);
                for (i = 0; i < zl; i++) srx_mat(&fsx,i,zpa[i]);
                if ((dbg > 1) && (fsx)) {
                    printf("exc:");
                    for (i = 32; i < 127; i++) printf("  %c ",i);
                    printf("\n");
                    for (i = 0; i <= fst; i++) {
                        printf("%3d ",i);
                        for (j = 32; j < 127; j++) printf("%3d ",fsx[i][j]);
                        printf("\n");
                    }
                }
                for (i = 0; i < vl; i++) {
                    /* fprintf(stderr,"%d %d %s\n",vl,i,vpa[i]); */
                    srx_mat(&fsm,i,vpa[i]); /* &vpa[i][(vpa[i][0]=='+')] */
                }
                if (dbg > 1) {
                    printf("fnd:");
                    for (i = 32; i < 127; i++) printf("  %c ",i);
                    printf("\n");
                    for (i = 0; i <= fst; i++) {
                        printf("%3d ",i);
                        for (j = 32; j < 127; j++) printf("%3d ",fsm[i][j]);
                        printf("\n");
                    }
                }
            }
        } else if (rl != vl) {
            printf("!! unmatched search-replace lists.");
            exit(6);
        }
        if (!(sw & _H_) && !(tim))
        {
            time(&tim);
            printf("------- %s",ctime(&tim));
            printf("------- list:   %s, %d words, %d exclusions.\n",words,vl,xl);
            printf("------- files:  %s; report:  lines / longest / scanned / hits\n\n",fspec);
        }
    }

    strcpy(fname,fn);

    if ((!fn[0]) || (!strcmp(fn,"stdin"))) {
        fp = stdin;
    } else {
        if (s = stat(fname,&sta)) {
            if (!(sw & _Q_)) fprintf(stderr,"\n!! stat error (%s:  %d)\n",fname,s);
            exit(s);
        }
        if (S_ISDIR(sta.st_mode)) {
            if (dr = opendir(fname)) {
                strcpy(ds,fname);
                strcat(ds,"/");
                while ((dn = readdir(dr)) != NULL) {
                    if (strcmp(dn->d_name,"..") && strcmp(dn->d_name,".")) {
                        strcpy(fn,ds);
                        strcat(fn,dn->d_name);
                        srx_read(fspec,fn,fct,cmnt,sw,wnd,lln,dbg,old,nsec);
                    }
                }
                closedir(dr);
            }
            return(0);
        }
        if (ftag[0]) {
#ifndef VMS
            if (!fnm_match(ftag,fn,0)) return(0);
#else
            if (!(strstr(fn,ftag))) return(0);
#endif
        }
        if ((ntag[0]) && (strstr(fn,ntag))) return(0);
        if ((nsec) && (old ? (sta.st_mtime > nsec) : (sta.st_mtime < nsec))) return(0);
        strcpy(ftime,ctime(&sta.st_mtime));
        ftime[24] = '\0';
        if (sw & _M_) m[1] = 'b';
        if ((fp = fopen(fname,m)) == NULL) {
            if (!(sw & _Q_)) fprintf(stderr,"!! can't open %s (%d)\n",fn,errno);
            exit(errno);
        }
    }

    /* allocate the context window; the 'hits' list is about
        reporting - if we report a line in context of one hit,
        we don't want to write it again in another, so we
        'turn it off' by setting its hits[] element;
    */
    assert((bb = (char **)calloc((wnd),sizeof(char **))) != NULL);
    for (i = 0; i < wnd; i++) {
        assert((bb[i] = (char *)calloc(1,lln+2)) != NULL);
    }
    assert((hits = (int *)calloc((wnd),sizeof(int *))) != NULL);

    /* generally, we don't care much about the content of comments
        in source files, so we are fairly aggressive about finding
        a set of comment tags for the file type - unless we've
        been told to include them;
    */
    if (!(sw & _C_)) srx_comment(fn,cmnt,cm,ct);
    if (cm[0]) {
        cmn = strlen(cm);
        ctn = strlen(ct);
    }

    if (dbg > 1) {
        printf("\n");
        printf("-- %-37s","Search words:");
        if (rl) printf("-- Replacement words:");
        printf("\n");
        for (i = 0; i < vl; i++) {
            printf("%-40s",vpa[i]);
            if (rl > i) printf("%s",rpa[i]);
            printf("\n");
        }
        if (xl) {
            printf("-- Exclusion wrappers:\n");
            for (i = 0; i < xl; i++) printf("-%s\n",xpa[i]);
        }
        if (zl) {
            printf("-- Peremptory exclusions:\n");
            for (i = 0; i < zl; i++) printf("~%s\n",zpa[i]);
        }
        printf("\n");
    }

    ln = hn = rn = mln = 0;
    b = bb[0];                                      /* read from the file into one of the */
    while (srx_fgets(b,lln,fp,sw)) {                /*   context buffers */
        eol = ((sw & _A_) ? 0 : 1);                 /* end-of-line is suppressed by '_Append' */
        szb = strlen(b);
        if (!rl && !(sw & _T_)) {
            fst = 0;
            memset(fsh,0,vl*sizeof(int));
        }
        /*
            scan each byte of the input line and port them to the 'rb'
            read buffer; the two issues are removal of comments to prevent
            false or noisy hits, and variously compressing the text to
            make hits more likely; we'll scan rb for hits, but we'll report
            them in the context of the bb[] array;
        */
        for (q = i = 0; (i < szb); i++) {
            /* trim comments:
              cksm srch and dff share this algorithm:
                comments may be c-style:  intro->terminator,
                    or c++ style:  intro->endofline;
                terminator-style comments may be nested to any reasonable depth, but the
                    delimiters must balance; we yell if they don't;
                    we don't care whether the compiler permits nesting;
                delimiters included in quoted strings within code will not be processed
                    as comment delimiters - be sure to clear q at eol;
            */
            if (cmn) {
                if ((!icx) && (b[i] == 34)) q = (1 & !q);
                if (!q)
                  while ((i < szb)
                    && (!strncmp(cm,&b[i],cmn) || !strncmp(ct,&b[i],ctn))) {
                    if (!strncmp(cm,&b[i],cmn)) {
                        icx++;
                        i += cmn;
                    } else if (!strncmp(ct,&b[i],ctn)) {
                        if (icx > 0) {
                            icx--;
                        } else if (ct[0] != '\n') {
                            printf("!! extra comment terminator at %d in:\n%s",i,b);
                            printf("!! %s\n",fname);
                        }
                        if (ct[0] == '\n') break; else i += ctn;
                    }
                }
            }
            if (b[i] == '\n') break;
            /*
                _Ignore case, be a_Ggressive, or _Append lines
            */
            if (!icx) {
                if ((sw & _G_) == _G_) {
                    if (isalnum(b[i])) rb[rp] = tolower(b[i]);
                } else if (sw & _A_) {
                    if (!isspace(b[i])) rb[rp] = b[i];
                    if ((b[i] == ';') || (b[i] == '#') || (rp > MAXLIN-2)) eol = 1;
                } else if (sw & _I_) {
                    rb[rp] = tolower(b[i]);
                } else {
                    rb[rp] = b[i];
                }
                rp++;
            }
        } /* eol */
        rb[rp] = '\0';                              /* make sure rb terminates */
        if (dbg) printf("%s\n",rb);
        if (mln < (mn = i)) mln = mn;               /* set the max line length */
        if (!rp && !(sw & _S_)) eol = 0;            /* process empty lines only for swaps */
        if (eol) {                                  /* if eol is suppressed, go get the next line */
            if ((invert) ^ ((vp = srx_scan(rb)) != NULL)) {
                if (!hn) {                          /* print a banner on the first hit, unless we're 'bare' */
                    if (sw & _F_) {
                        if (!(sw & _N_)) printf("%s\n",fname);
                    } else if (!(sw & _B_)) {
                        printf("+****** %s %s ",fname,ftime);
                        if (cm[0])       printf("%s comments",cm);
                        if (ct[0] > ' ') printf(" %s",ct);
                        printf("\n");
                    }
                }
                hn++;                               /* score the hit */
                if (!(sw & _W_)) {
                    ps = wnd-1;                     /* set the postscript context lines */
                    k = ((ln < wnd) ? ln+1 : wnd);  /*  if we haven't read a full window of lines, */
                    for (i = k-1; i >= 0; i--) {    /*  report what we have behind us; */
                        c = (i ? ' ' : '>');
                        j = (ln - i) % k;
                        if (!hits[j]) {             /* report each line in the window only once */
                            if (!(sw & _B_) && !invert) printf(hmsk,(ln-i)+1,vp,c);
                            if (sw & _M_)
                                printf("%s\n",rb);
                            else
                                printf("%s",bb[j]);
                            hits[j] = 1;
                        }
                    }
                }
            } else if (ps > 0) {                    /* 'postscript' opens the context window a little wider */
                if (!(sw & _B_)) printf(pmsk,ln+1,"");          /* by printing 'wnd-1' lines */
                printf("%s",bb[bn]);                            /* after the hit; */
                hits[bn] = 1;                                   /* the cosmetics are designed */
                ps--;                                           /* for a little eye-relief */
                if (!(ps) && !(sw & _B_)) printf("-------\n");
            }
            rb[(rp = 0)] = '\0';
            rn++;
        }
        ln++;                                       /* incrementing the line count is obvious, but then */
        bn = (ln % wnd);                            /* we need to rotate the window buffer one slot */
        b = bb[bn];                                 /* by taking the modulus of the line number, */
        hits[bn] = 0;                               /* then clear the 'hits' flag, */
    } /* eof */
    if (fp != stdin) fclose(fp);

    /* print a trailer if we're not running bare, then kick all the
        stats up to globals
    */
    if ((hn) || (sw & _Z_)) {
        if (!(sw & _H_)) {
            printf("-****** %s %d / %d / %d / %d\n\n",fname,ln,mln,rn,hn);
        } else if ((sw & _N_) && (hn == 0)) {
            printf("%s\n",fname);
        }
    }
    lines       += ln;
    scanned     += rn;
    hitcount    += hn;
    if (longest < mln) longest = mln;
    files++;
    return(1);
}

char *srx_fgets(char *b, int lln, FILE *fp, int sw) {
    int i = 0;
    int c = 0;

    b[0] = '\0';
    if (feof(fp)) return(NULL);
    for (i = 0; i < lln; i++) {
        c = fgetc(fp);
        if ((sw & _X_) && (c == 0)) c = fgetc(fp);
        if (c == EOF) break;
        if (((c >= 0x20) && (c <= 0x7e))
        ||  ((sw & _U_) && (c >= 0xa0) && (c <= 0xfe))
        ||  (c == '\t')
        ||  (c == '\n')
        ) {
            b[i] = 0xff & c;
        } else {
            b[i] = '.';
        }
        if (c == '\n') break;
    }
    b[i+1] = '\0';
    return(b);
}

/*****************************************************************
 srx_scan:  scan the buffer for hits; pass off the real work to
    srx_hit; we have four variations:
        srx_scanf - the usual, fast version, uses the state machine;
        srx_scant - turtle version, as fast or faster than scanf
                    for 1 or 2 strings;
        srx_scanr - search/replace;
        srx_scanp - search only at the specified position
        srx_scanx - search and exclude at the specified position

    the business with the '+' prefix is an 'and' list:  to be
        recorded as a 'hit', all the members of a contiguous
        sequence which start with '+' have to be found in the
        buffer;

    a word-list might look like this (without the comments):

mercury     - any of the planets generates a hit by itself
venus
earth
mars
jupiter
saturn
uranus
neptune
pluto
-parvenus   - unless it's not really a planet name
-ditmars
-marshal
-dearth
-hearth
-earthly
+io         - but moons all have to be in one buffer or it's not a hit
+europa
+titan
+charon
~sol        - and stars generate peremptory exclusions
~betelgeuse
~sirius
~deneb

*/
char *srx_scanf(char *rb)
{
    int     i           = 0;            /* iterators */
    char    *vp         = NULL;
    int srx_match(char *buffer, int **machine, int immed);

    /* check the peremptory exclusion list first
    */
    if (srx_match(rb,fsx,1)) return(NULL);

    srx_match(rb,fsm,0);

    for (i = 0; i < vl; i++) {
        if (vpa[i][0] == '+') {
            if (!fsh[i]) return(NULL);
        } else {
            if (fsh[i]) return(vpa[i]);
        }
    }
    return(vp);
}
char *srx_scant(char *rb)
{
    int     i           = 0;            /* iterators */
    char    *vp         = NULL;
    char *srx_hit(char *rb,char *vp);
    char *srx_and(char *rb, int *k);

    /* check the peremptory exclusion list first
    */
    for (i = 0; i < zl; i++) if (strstr(rb,zpa[i])) return(NULL);

    for (i = 0; i < vl; i++) {
        if (vpa[i][0] == '+') {
            if (vp = srx_and(rb,&i)) break;
        } else {
            if (vp = srx_hit(rb,vpa[i])) break;
        }
    }
    return(vp);
}
char *srx_scanp(char *rb)
{
    int     i           = 0;            /* iterators */
    char *srx_scanx(char *rb);

    if (srx_scanx(rb)) return(NULL);
    for (i = 0; i < vl; i++) {
        if (!strncmp(&rb[psn],vpa[i],strlen(vpa[i]))) break;
    }
    if (!vl || (i < vl)) {
        hitcount++;
        if (vpa[i]) return(vpa[i]);
        else        return(xpa[0]);
    }
    return(NULL);
}
char *srx_scanx(char *rb)
{
    int     i           = 0;            /* iterators */

    for (i = 0; i < zl; i++) {
        if (!strncmp(&rb[psx],zpa[i],strlen(zpa[i]))) return(zpa[i]);
    }
    return(NULL);
}

/*****************************************************************
 srx_scanr, formerly
 srx_replace:  a late-entry feature, silly, but sometimes handy;
    any number of replacement strings can be in the list, as long as
    they balance search strings; the terminator is non-negotiable:
    '?', so an entry looks like:

?fixme?donefixed

    the search strings will behave exactly like ordinary hit-checks,
    honoring any exclusions that are in the word-list; however,
    be careful not to include any search words that are not part of
    the replacement list;

    this list will work fine:

?fixme?donefixed
-dontfixme
-norfixmeeither
~dontconsiderthis

    but it breaks if you add:

searchforthis

    because the search-replace pairs are no longer balanced;

    srx_replace prints its own output, and updates the hitcount global;
    the caller, srx_read, doesn't know about multiple hits per line, and
    it has very different rules for output display;
    psn is unused;
*/
char *srx_scanr(char *rb)
{
    int     i           = 0;            /* iterators */
    int     j           = 0;            /*     "     */
    char    *vp         = NULL;
    char    *rp         = NULL;
    char    t[MAXLIN]   = "";
    char    tmp[MAXLIN] = "";
    char *srx_hit(char *rb,char *vp);

    //printf("- %s\n",rb);
    for (i = 0; i < zl; i++) if (strstr(rb,zpa[i])) {
        printf("%s\n",rb);
        return(NULL);
    }
    strcpy(t,rb);
    for (i = 0; i < vl; i++) {
        if (strcmp(vpa[i],rpa[i])) {
            rp = t;
            while (vp = srx_hit(rp,vpa[i])) {
                vp = strstr(rp,vpa[i]);
                j = strlen(tmp) + (vp-rp);
                strncat(tmp,rp,(vp-rp));
                tmp[j] = '\0';
                strcat(tmp,rpa[i]);
                //printf("1 %s\n",tmp);
                rp = &vp[strlen(vpa[i])];
                hitcount++;
            }
            if (tmp[0]) {
                strcat(tmp,rp);
                //printf("2 %s\n",tmp);
                strcpy(t,tmp);
                tmp[0] = '\0';
            }
        }
    }
    printf("%s\n",t);
    return(NULL);
}

/*****************************************************************
 srx_and:  keep trying until we get to the end of the and-list or a match fails;
    either way, update the string counter to the end of the and-list;
*/
char *srx_and(char *rb, int *k)
{
    int     i           = 0;            /* iterators */
    char    *vp         = NULL;
    char *srx_hit(char *rb,char *vp);

    vp = srx_hit(rb,&vpa[*k][1]);
    for (i = (*k + 1); (i < vl) && (vpa[i][0] == '+'); i++) {
        if ((vp) && !srx_hit(rb,&vpa[i][1])) vp = NULL;
    }
    *k = i;
    return(vp);
}

/*****************************************************************
 srx_hit:  a worker bee; simple enough:  consider only one search string,
    and negate the result if it's found within an exclusion string;
    if we exclude what we found, jump past the found string and try again;
    in trying exclusion strings, we can start searching a word-length
    behind the found string, if we're that far into the buffer;
*/
char *srx_hit(char *rb,char *vp)
{
    int     i           = 0;            /* iterators */
    int     j           = 0;            /*     "     */
    int     s           = 0;
    char    *found      = NULL;
    char    *exclude    = NULL;
    char    *tryf       = NULL;
    char    *tryx       = NULL;

    tryf = rb;
    while (found = strstr(tryf,vp)) {
        s = 1;
        for (i = 0; i < xl; i++) {
            tryx = (tryf - rb > strlen(xpa[i]) ? tryf - strlen(xpa[i]) : rb);
            while ((exclude = strstr(tryx,xpa[i])) && (exclude <= found)) {
                if (&exclude[strlen(xpa[i])] >= &found[strlen(vp)])
                    s = 0;
                if ((s) && (strlen(tryx) > strlen(xpa[i])))
                    tryx += strlen(xpa[i]);
                else
                    break;
            }
        }
        if (s) return(vp);
        if (strlen(tryf) > strlen(vp))
            tryf += strlen(vp);
        else
            break;
    }
    return(NULL);
}

/*****************************************************************
 srx_words:  find & open the word list, and add all the words to the list(s)
    (srx_word does the dirty work, he'll sort out which list each word belongs to)
*/
int srx_words(char *fn, int sw)
{
    int     i           = 0;            /* iterators */
    int     j           = 0;
    int     r           = 0;
    FILE    *fp         = NULL;
    char    txt[256]    = "";
    int     srx_word(char *, int sw);

    if ((!fn[0]) || (!strcmp(fn,"stdin"))) {
        fprintf(stderr,"_Search string(s):\n");
        fp = stdin;
        strcpy(words,"stdin");
    } else {
        if ((fp = fopen(fn, "r")) == NULL) {
            fprintf(stderr,"!! can't open %s (%d)\n",fn,errno);
            exit(errno);
        }
        strcpy(words,fn);
    }
    while (fgets(txt,255,fp) != NULL) {
        if ((txt[0] < ' ') && (txt[0] != 9)) {
            break;
        } else {
            txt[strlen(txt)-1] = '\0';
            r |= srx_word(txt,sw);
        }
    }
    if (fp != stdin) fclose(fp);
    return(r);
}

/*****************************************************************
 srx_word:  the guts of the srx_words loop;
    check the leadin character, and decide whether to put the word into one of
    the exclusion lists, or the search list;
    the real magic of this routine is the allocation scheme; because it
    reallocs the lists on the fly, we can add hundreds of search strings
    to the tables; the run-time penalty is very small - a typical word-list
    will go through a million lines of code per minute; allocating 8 slots at a
    time is purely arbitrary, 1 or 2 would probably do as well
    since this only happens once at startup;
*/
int srx_word(char *t, int sw)
{
    int     i           = 0;            /* iterators */
    int     j           = 0;            /*     "     */
    int     r           = 0;
    static  int m       = 0;            /* max word-length */
    char    hmskmsk[32] = "%%7d %%-%ds-%%c";
    char    pmskmsk[32] = "%%7d %%-%ds  ";
    char    *vp         = NULL;         /* pointer to add to whichever list */

    /* fprintf(stderr,"%s\n",t); */
    if (sw & _I_)
        for (j = 0; j < strlen(t); j++)
            t[j] = tolower(t[j]);
    assert((vp = (char *)calloc(1,strlen(t)+1)) != NULL);
    switch (t[0])
    {
        /* exclusion:  if the string found is a substring of this,
            don't count it as a hit
        */
        case '-':
        {
            strncpy(vp,&t[1],strlen(t)-1);
            if (!(xl%8)) {
                assert((xpa = (char **)realloc(xpa,(xl+8)*sizeof(vp))) != NULL);
            }
            xpa[xl++] = vp;
            r = _T_;
            /* fprintf(stderr,"%-5d %s\n",xl,vp); */
            break;
        }
        /* search-replace:  must occur in pairs, delimited as:  ?oldword?newword
        */
        case '?':
        {
            strncpy(vp,&t[1],strlen(t)-1);
            if (!(rl%8)) {
                assert((rpa = (char **)realloc(rpa,(rl+8)*sizeof(vp))) != NULL);
            }
            for (i=0; (vp[i]); i++) if (vp[i]=='?') { vp[i++]='\0'; break; }
            rpa[rl] = &vp[i];
            for (i=j=0; (rpa[rl][j]); j++) {
                if (rpa[rl][j] == '\\') {
                    switch (rpa[rl][++j]) {
                        case 'a': rpa[rl][i] = '\a'; break;
                        case 'n': rpa[rl][i] = '\n'; break;
                        case 'r': rpa[rl][i] = '\n'; break;
                        case 't': rpa[rl][i] = '\t'; break;
                        default:  i++; break;
                    }
                }
                i++;
            }
            rpa[rl][i] = '\0';
            rl++;
            if (!(vl%8)) {
                assert((vpa = (char **)realloc(vpa,(vl+8)*sizeof(vp))) != NULL);
            }
            vpa[vl++] = vp;
            r = _R_;
            /* fprintf(stderr,"%-5d %s\n",rl,vp); */
            break;
        }
        /* peremptory exclusion:  if this string occurs in a line, don't scan the line
        */
        case '~':
        {
            strncpy(vp,&t[1],strlen(t)-1);
            if (!(zl%8)) {
                assert((zpa = (char **)realloc(zpa,(zl+8)*sizeof(vp))) != NULL);
            }
            zpa[zl++] = vp;
            r = _T_;
            /* fprintf(stderr,"%-5d %s\n",zl,vp); */
            break;
        }
        /* 'escape' char, ','
        */
        case ',':
        {
            strncpy(vp,&t[1],strlen(t)-1);
            if (!(vl%8)) {
                assert((vpa = (char **)realloc(vpa,(vl+8)*sizeof(vp))) != NULL);
            }
            vpa[vl++] = vp;
            /* fprintf(stderr,"%-5d %s\n",vl,vp); */
            break;
        }
        case '+':
        {
            r = _T_;    /* set turtle mode, and fall through */
        }
        /* find this string
        */
        default:
        {
            strncpy(vp,t,strlen(t));
            if (!(vl%8)) {
                assert((vpa = (char **)realloc(vpa,(vl+8)*sizeof(vp))) != NULL);
            }
            vpa[vl++] = vp;
            /* fprintf(stderr,"%-5d %s\n",vl,vp); */
            break;
        }
    }
    if (m < strlen(vp)) {
        m = strlen(vp);
        sprintf(hmsk,hmskmsk,m);
        sprintf(pmsk,pmskmsk,m);
    }
    return(r);
}

/*****************************************************************
 srx_match:  this elegant piece of work is from Moishe Halibard and
    Moshe Rubin, "A Multiple Substring Search Algorithm", CUJ June 2002.

    The idea is to build a finite-state-machine for a list of substrings,
    then hit the machine with a buffer of text, one character at a time.
    As long as the state is 0, there is no match.  When we get a match,
    the state will keep incrementing until we have a mismatched character,
    which will return 0, or end-of-substring, which returns a negative.

    If it's a 0, go back to the starting posit, and resume searching;
    if it's negative, we have to process a 'hit':

    In immediate mode, we just call it a hit and bail; otherwise, we might
    be looking for all the members of an 'and' list (see notes on _scan),
    so we mark the fsh hits array, and hang in there until the end of the
    buffer.
*/
int srx_match(char *buffer, int **machine, int immed)
{
    int     i           = 0;            /* iterator */
    int     state       = 0;            /* the 'current' state */
    int     interim     = 0;            /* the prior state */
    int     posit       = 0;            /* buffer position when the state was last 0 */

    if (!machine) return(0);

    for (i = 0; (buffer[i]); i++) {
        if (!(interim = state)) posit = i;
        if (!(state = machine[state][buffer[i]])) {
            if (interim) i = posit;
        } else {
            if (state < 0) {
                if (immed) return(state);
                fsh[(-state) - 1] = state;
                state = 0;
            }
        }
    }
    return(0);
}

/*****************************************************************
 srx_mat:  create & load the finite-state matrix

    This is nasty to debug (depending on the strength of your C), but
    the execution is fairly easy:

    For each search string we're passed, expand the machine array by one
    line per character, 256 elements per line (256 is vast overkill, since
    we're only dealing with printable characters, but memory is cheap,
    and I hate access violations).

    For each new character in the substring, increment the global state
    and pop the new state value into the array at (priorstate,charvalue).
    When we get to end-of-string, put the negative string number in the
    (priorstate,charvalue).  That will give the reader proc (srx_match)
    an indication of a full match.

    The sm<->sma fiddling is because I want two machines, one for
    exclusion strings, and one for search strings.
*/
void srx_mat(int ***sma, int snum, char *s)
{
    int     i           = 0;            /* iterator */
    int     j           = 0;            /* iterator */
    int     n           = 0;            /* length of s */
    int     t           = 0;            /* temporary state */
    int     **sm        = NULL;

    sm = *sma;
    n = strlen(s);
    if (!snum) fsz = fst = 0;
    if (sm) {
        assert((sm = (int **)realloc(sm,(fsz+n)*sizeof(int *))) != NULL);
        for (i = 0; i < n; i++) sm[fsz+i] = NULL;
    } else {
        assert((sm = (int **)calloc(n,sizeof(int *))) != NULL);
    }
    /* printf("sm %p  max t %5d  s %s",sm,fsz+n,s); */
    for (i = 0; (i < n) && (t >= 0); i++) {
        /* printf(" %d %d %c ",i,t,s[i]); */
        if (i < (n-1)) {
            if ((sm[t]) && (sm[t][(s[i])])) {
                /* printf(" sm[t][(s[i])] %d ",sm[t][(s[i])]); */
                t = sm[t][(s[i])];
            } else {
                /* printf(" t %d  sm[t] %p  fst %d  s[i] %d ",t,sm[t],fst,s[i]); */
                if (!sm[t]) {
                    assert((sm[t] = (int *)calloc(256,sizeof(int))) != NULL);
                }
                /* printf(" smtsi %d fst %d ",sm[t][(s[i])],fst); */
                sm[t][(s[i])] = ++fst;
                t = fst;
                /* printf(" t %d ",t); */
            }
        }
    }
    fsz += n;
    /* printf(" %d\n",t); */
    if (t >= 0) {
        if (!sm[t]) {
            assert((sm[t] = (int *)calloc(256,sizeof(int))) != NULL);
        }
        sm[t][(s[n-1])] = -(snum+1);
    }
    *sma = sm;
}

/*****************************************************************
 srx_comment:
    if we're given an explicit comment delimiter set (in comnt), then split it and return;
    if not, pick out the file extension from fn, search for it in the list, and use the matching
    delimiter set;
    if we don't find a match in the list, use the environment variable SRCH_COMMENT;
    or give up;
*/
void srx_comment(char *fn, char *comnt, char *cm, char *ct)
{
    int     i           = 0;            /* iterator */
    int     j           = 0;            /* iterator */
    char    cmnt[10]    = "";           /* the comment delimiters we're going to return */
    char    ext[20+1]   = "";           /* file extension */
    typedef struct {
            char        *ex;
            char        *dlm;
    } cmdel;
    cmdel   cl[]        = {
                        {".bas"  ,"!"}  /* comment delimiter list */
                       ,{".com"  ,"!"}
                       ,{".def"  ,"(* *)"}
                       ,{".mod"  ,"(* *)"}
                       ,{".c"    ,"/* */"}
                       ,{".h"    ,"/* */"}
                       ,{".cxx"  ,"//"}
                       ,{".hxx"  ,"//"}
                       ,{".cpp"  ,"//"}
                       ,{".hpp"  ,"//"}
                       ,{".java" ,"//"}
                       ,{".tcl"  ,"#"}
                       ,{".tk"   ,"#"}
                       ,{NULL    ,NULL}
                        };

    cm[0] = ct[0] = '\0';
    if (comnt[0]) {
        strcpy(cmnt,comnt);
    } else {
        for (i = strlen(fn); (i) && (fn[i] != '.'); i--) ;
        if (i) {
            for (j = i; ((j-i) < 20) && (fn[j]) && (fn[j]!=';'); j++) ext[j-i] = tolower(fn[j]);
            ext[j-i] = '\0';
            for (i = 0; (cl[i].ex); i++) {
                if (!strcmp(ext,cl[i].ex)) {
                    strcpy(cmnt,cl[i].dlm);
                    break;
                }
            }
        }
    }
    if (cmnt[0]) {
        for (i = 0; (cmnt[i]); i++)
            if ((cmnt[i] == ' ')
            ||  (cmnt[i] == '-')
            ||  (cmnt[i] == '.')
            ||  (cmnt[i] == '_'))
                break;
        if (cmnt[i]) {
            cmnt[i++] = '\0';
            strcpy(ct,&cmnt[i]);
        } else {
            ct[0] = '\n';
            ct[1] = '\0';
        }
        strcpy(cm,cmnt);
    }
}

/*****************************************************************
*/
void srx_help(int sw)
{
    printf(" srch -abCgimq0 -l50 -d1 -c/*_*/ ,str1 ,str2 . . . file.1 file.2 . . .\n");
    printf(" srch -help\n");

    if (!sw) exit(0);

    printf("\n");
    printf(" SRCH needs a search list of words, an optional output file, and input file(s).\n");
    printf("  The wordlist may be a file, 1 word per line, or command-line parameters.\n");
    printf("  Prefix a wordlist filename with '{', the output file with '}'.\n");
    printf("\n");
    printf(" Optional switches, anywhere prior to the files to search:\n");
    printf("  -! invert the sense of a 'hit'\n");
    printf("  -a Append extended lines before searching/excluding (.c)\n");
    printf("  -b Bare output, just text w/hits\n");
    printf("  -c include Comments in scan\n");
    printf("  -C/*_*/ -C! comment delims, to add to or override defaults\n");
    printf("  -d1 Debug, show scanned text\n");
    printf("  -d2 Debug, show word-lists and scan buffer\n");
    printf("  -F<file> search File\n");
    printf("  -g aGgressive search, no space, no case, no punctuation\n");
    printf("  -h this help, -H more help\n");
    printf("  -i case-Insensitive\n");
    printf("  -l#### set line-Length (binaries)\n");
    printf("  -m Machine-code (binary/exe/obj) file scan\n");
    printf("  -n No, there was no hit:  filenames only\n");
    printf("  -O<file> re-direct Output\n");
    printf("  -p## look only in position ##\n");
    printf("  -q Quiet, file names and hits\n");
    printf("  -r Recursive globbing of filename(s)\n");
    printf("  -S<file> Search-word list\n");
    printf("  -t Turtle mode, don't use the fsm to optimize searching\n");
    printf("  -u(-8) don't skip high-bit chars\n");
    printf("  -w## window size, for hit context\n");
    printf("  -x## exclude at position ##\n");
    printf("  -X wide chars, skip nulls\n");
    printf("  -y Yes there was a hit:  filenames only\n");
    printf("  -z report names of files scanned w/zero hits\n");
    printf("  -*<tag> search only files with names including <tag>\n");
    printf("  -#<tag> search only files with names excluding <tag>\n");
    printf("  -- end of switches and search-strings (only filenames follow)\n");
    printf("  n.b.:  if the file list includes a directory,\n");
    printf("      it will be searched recursively, regardless of '-r' setting;\n");
    printf("\n");
    printf("  (don't leave space between a switch and its value)\n");
    printf("\n");
    printf("  The word-list is plain text, one 'word' to a line, but words may contain\n");
    printf("    spaces or punctuation. \n");
    printf("  Combinations (match=and) are prefixed by a '+' on each word.\n");
    printf("  Exclusion words are signalled by a leading '-'.  Exclusions are used\n");
    printf("    to prevent false 'hits' by surrounding a search-word:\n");
    printf("    'date' will hit on 'update' unless '-update' is loaded in the list.\n");
    printf("  Peremptory exclusions (~word):  if found, exclude the line from the scan.\n");
    printf("  Searching for nonsense with switches '-bd1' outputs a complete source file\n");
    printf("  stripped of all comments.\n");
    printf("  Replacement wordlist syntax:  ?find string?replace str\n");
    printf("  Command-line wordlist, prefixing each word with a comma:\n");
    printf("    srch ,word1 ,word2 *.files\n");
    printf("  or equivalently, terminate the list with '--':\n");
    printf("    srch word1 word2 -- *.files\n");
    printf("\n");
    printf("  Syntax:  srch {WORD.LIS }OUT.LIS INFILES.*\n");
    printf("    alt.:  srch ,word \",-xwordx\" }out.txt infile*.*\n");
    printf("\n");
    printf("    v.%s, %s %s\n",VERSION,BUILD,AUTHOR);
    printf("\n");

    if (sw < 2) exit(0);

    printf(" Comment processing:\n");
    printf("\n");
    printf("  SRCH tries to exclude comments from searches of the file types\n");
    printf("  that it recognizes, which are source files for the C family,\n");
    printf("  BASIC, Modula-2, Tcl, and DCL.  Other comment delimiters can\n");
    printf("  be specified on the command line via the -c switch, but only\n");
    printf("  one set of delimiters can be in effect at a time.  Support for\n");
    printf("  COBOL and Fortran fixed-column comment tags has fallen away.\n");
    printf("  Comment-pairs, like C's /* */ can follow -c by separating the\n");
    printf("  leader and trailer with underscore:  -c/*_*/\n");
    printf("\n");
    printf(" String replacement:\n");
    printf("\n");
    printf("  Search/replace is not exercised much, but the syntax is simple:\n");
    printf("  all words must appear in pairs, with a leading accent-grave:\n");
    printf("\n");
    printf("  ?fixme?iamfixed\n");
    printf("\n");
    printf("  The search strings will behave exactly like ordinary hit-checks,\n");
    printf("  honoring any exclusions that are in the word-list; however,\n");
    printf("  be careful not to include any search words that are not part of\n");
    printf("  the replacement list.  This list will work fine:\n");
    printf("\n");
    printf("  ?fixme?donefixed\n");
    printf("  -dontfixme\n");
    printf("  -norfixmeeither\n");
    printf("  ~dontconsiderthis\n");
    printf("\n");
    printf("  But it breaks if you add:\n");
    printf("\n");
    printf("  justfindthis\n");
    printf("\n");
    printf("  because the search-replace pairs are no longer balanced.  The\n");
    printf("  output is written to stdout, so you can either watch the text\n");
    printf("  scroll away, or re-direct it to a new file.\n");
    printf("\n");
    printf(" Machine code:\n");
    printf("\n");
    printf("  Searching executables or object files for strings is sometimes\n");
    printf("  handy, for ascertaining library module names or error messages,\n");
    printf("  for example.  Most of the file will be unprintable junk, so\n");
    printf("  only text is scanned, and the scan buffer is what prints when\n");
    printf("  a hit is detected.  Since the scan buffer is a fixed length\n");
    printf("  (there being no reliable record terminators in an object file)\n");
    printf("  a string might be split across buffers and missed, so it pays\n");
    printf("  to try one or two additional scans with arbitrary changes in\n");
    printf("  the buffer length, -l85 in one search and -l112 in another,\n");
    printf("  for example.\n");
    printf("\n");
    printf(" Position search:\n");
    printf("\n");
    printf("  Cheap database retrieval. All search-words will be tried, but only\n");
    printf("  at the character position -p#.  Likewise, exclusion words ,~<word>\n");
    printf("  will filter records if any of the exclusion words occur at position\n");
    printf("  -x#.  The exclusion words are peremptory.\n");
    printf("\n");
    printf(" Turtle mode:\n");
    printf("\n");
    printf("  The original design of SRCH used a brute-force approach of\n");
    printf("  applying strstr to the text, looping through the list of words.\n");
    printf("  This works, but it does get slower as the list gets longer,\n");
    printf("  and it might matter if you have a few million lines to search\n");
    printf("  for a hundred or so words.\n");
    printf("\n");
    printf("  A more elegant approach comes from Moishe Halibard and Moshe Rubin,\n");
    printf("  in their article \"A Multiple Substring Search Algorithm\", CUJ June 2002.\n");
    printf("  The algorithm calls for a state machine built from the wordlist, and\n");
    printf("  each character of text is applied to the machine so that redundant\n");
    printf("  searches are minimized.  I left the old algorithm in because\n");
    printf("  superstring exclusion and search-replace can't use this.  The fsm runs\n");
    printf("  by default in normal multi-string searches, but you can turn it off\n");
    printf("  via the -t switch.  If you want to see what's in the fsm, search a\n");
    printf("  small file with -d2 and a sizeable search-list.\n");
    printf("\n");

    exit(0);
}
