/* SRT.c - basic sort program, multiple keys w/key types (str, int, date)

    Another sort?  Well, aside from getting some exercise, there are still
    a couple of practical reasons for writing a new general-purpose sort: 
    most of the 'sorts' available don't come with source, so it's hard to
    find out why the last command didn't work the way you thought it
    would; most of them have arbitrary restrictions on file size, file
    type, record size, number of key fields, size of keys, or whatever,
    which are seemingly imposed just because the programmer made an array,
    and it can't be enlarged; the sorts that come with DOS, NT, VMS, and
    Unix work, but they aren't portable; third-party sorts may or may not
    be portable, but there will be a purchase and/or license fee for them,
    and you probably won't have source, which brings us back to the first
    argument.  So here's a free, portable sort with no arbitrary
    restrictions.  Just don't re-sell it, and if you change it, distribute
    the source with the binary. 

    Actually, SRT is an interface to a sort:  the qsort that comes in the
    Standard Run-time Library.  I wanted to play with the qsort function
    and arrays of pointers, and this was the result.  Therefore, your
    performance will vary, depending on what compiler and OS you use.  I
    may get around to trying some other sort algorithms eventually, but
    for now, this is a practical application. 

    As of 2005/08/15:

    Thanks to Matthew Aitkenhead and Mark Richards, we have a new algorithm
    to play with:  Ordersort.  The article appears in the Sep 2005 C Users
    Journal, from which I scrounged the implementation below.  Ordersort
    supersedes the call to qsort, unless the command-line specifies '-q'. 
    (Mon Aug 15 12:55:58 2005 227)

    If there are restrictions on file size or string length, I didn't put
    them there (or not deliberately).  32766 is the limit on record input,
    and it's a pretty common limit, because it's close to where a signed
    short int rolls over, and negative numbers get confused with EOF in
    some libraries.  Physical memory or quotas will blow when multi-
    megabyte files are read into memory, but that's not an 'arbitrary'
    limit, except in that I didn't implement multiple scratch files.
    If you hit a memory limit, try a -k (key) or -l (large file) switch.

    Keys:  not very interesting if you just want to 'sort <a >b', but
    multiple keys can be very useful for reports and such.  A key has five
    attributes in SRT:  key offset, record offset (position), direction
    (reverse or normal), data type and size (length in bytes). 

        The key offset is implicit, calculated from the order and length
        of the preceeding keys.

        'Position' is the offset from the beginning of the record, so it
        starts at 0.  This is the default, if you don't provide a switch
        '-p##'. 

        Any of the sort keys can be '-r'eversed.  There is no restriction
        on mixing ascending (default) and descending keys.

        Data 'Type' allows for different types of sort, including lexicals,
        integers, and century-windowing.  Type is specified by a '-t'
        switch, with a suffix for the type signal of 's' 'i' 'd' 'm' or 'l'. 

            The default type is 's'tring, or a simple ASCII comparison. 

            Numbers that aren't justified don't sort right, unless you
            declare a key (has to have a length) of type 'n'.

            If the file has binaries, 'i'nteger comparisons may be useful,
            but they haven't been tested very well,

            Y2k fans will like 'd'ate and 'm'mddyy, because they compare
            6-byte dates correctly, with a 00-09 > 90-99 century split. 
            If you have to sort non-compliant y2k data for testing, this
            data type pays for the whole program.

            Dictionary comparisons are tricky, because the definition is
            either complex or vague.  SRT has one simple-minded rule for
            'l'exical comparisons:  skip anything that isn't a letter, and
            compare the rest; if it matches, do a string comparison.  Not
            very sophisticated, but it works for me. 

            Other data types (i.e., sort patterns) may be added later,
            probably before I work on another sort algorithm. 

        'Size' has to be the last switch in a key.  Position or Type can
        be in either order, but -s## comes last so the parser doesn't get
        confused by multiple keys. 

    'Binary' files don't behave consistently across different OSs.  VMS
    can have variable-length binary records, but the library output is so
    slow that I opted for a fixed-length record format to keep SRT
    portable and fairly quick.  You *must* define binary output files
    (empty) before you run VMS SRT, or the output will not match the
    input format.  On other systems, the output can be generated on the
    fly.  In either case, '-b##' requires a suffix of the fixed record
    length, and each key must have an implied or specified size.

    'Large' files are a subjective definition, but if you suspect that
    your available virtual memory can't hold the entire file, with about
    5% to 10% overhead, try '-k' or '-l'.  A '-k'ey sort is a little
    slower than a memory-bound sort, because it has to read the input
    twice, but it only stores the keys in ram, so it may use a lot less
    space than the whole file.  A '-l'arge sort is very slow, because
    every comparison of records will go back to the disk.  But the effect
    is mitigated by caching on most systems, and you shouldn't run out of
    memory unless a file is huge.  I don't intend to add anything more
    sophisticated to optimize SRT for large input, such as multiple work
    files. 

    earlroosa@yahoo.com

    v1.0 - Wed Oct  7 13:31:01 1998
    v1.1 - binary data, large files, lexical strcmp
    v1.2 - fn_cmp[], reverse, stat-info
    v1.3 - key sort, char-i/o, no vbl-bin files
    v2.0 - added Ordersort (fn_order) Mon Aug 15 12:56:51 2005 227

*/

#define VERSION "2007267"
#define BUILD "Mon Sep 24 14:49:55 2007 267"
#define AUTHOR "E.Roosa"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#ifdef VMS
#include <unixio.h>
#endif
#ifdef _MSC_VER
#include <io.h>
#if    _MSC_VER > 1000
#define _LREC_ 32766
#else
#define _LREC_ 10000
#endif
#else
#define _LREC_ 32766
#endif
#include <math.h>

#ifdef __linux__
#define MAKE_SRT "gcc -dn -s -O2 -o srt srt.c -lm"
#else
#ifdef VMS
#define MAKE_SRT "cc srt; lin srt"
#else
#ifdef __WIN32__
#define MAKE_SRT "gcc -mno-cygwin -O2 -dn -s -o srt.exe srt.c -lm"
#else
#define MAKE_SRT "cc -dn -s -O2 -o srt srt.c"
#endif
#endif
#endif

/* array subscript values for the key data types (ky[n]->type): */
#define         K_STR 0
#define         K_INT 1
#define         K_DT6 2
#define         K_MDY 3
#define         K_NUM 4
#define         K_LEX 5

char            xx[_LREC_] = "";    /* uniq buffer */
char            tx[_LREC_] = "";    /* input buffer */
char           *txt;                /* pointer to same */
long            ip;                 /* record location */
int             ksz = 0;            /* total key size, all segments */
int             rsz = 0;            /* current record size */
int             rnm = 0;            /* record number (for qsort) */
int             lrg = 0;            /* large file? use disk (slow) */
int             kys = 0;            /* large-ish file? key sort */
int             bin = 0;            /* binary data? */
int             dbg = 0;            /* debug? (show tags & pointers) */
int             inf = 0;            /* verbose? (show time & version) */
int             qst = 0;            /* ordersort/qsort */
int             unq = 0;            /* uniq output */
char            Dcd = '9';          /* windowing decade, for yymmdd */

typedef struct KY
{
        int     pos;                /* record buffer position (fr 0) */
        int     siz;                /* segment size in bytes */
        int     typ;                /* data type (0-4) */
        int     rev;                /* reverse? (descending sort) 1 / -1 */
        int     tos;                /* tag position (key sort) */
} KY;
KY    **ky = NULL;                  /* key array */
int     kn = 0;                     /* number of key segments */

typedef struct TG
{
        char   *tag;                /* tag or text of record */
        long    ptr;                /* file location (fr tell) */
} TG;
TG    **tg = NULL;                  /* tag array */
int     tn = 0;                     /* current record (tag) number */

FILE   *fip = NULL;                 /* input file pointer */
FILE   *fop = NULL;                 /* output file pointer */

int fn_makeky(int kn,int kp,int ks,char kd,int kr);
int fn_maketg(int tn, long *p);
int fn_open(char *fi, char *m, FILE **fp);
int fn_read(long *p,int w,char *t);
int fn_write();
int fn_time();
int fn_compare(const void *tg1, const void *tg2);
void fn_abort(char *msg, char *vbl);
int fn_mdycmp(char *d1, char *d2, int n);
int fn_dt6cmp(char *d1, char *d2, int n);
int fn_intcmp(void *i1, void *i2, int n);
int fn_lexcmp(char *d1, char *d2, int n);
int fn_strcmp(char *d1, char *d2, int n);
int fn_numcmp(char *d1, char *d2, int n);
void fn_help();
void fn_order(TG **A, int N);

int main(int argc,char *argv[])
{
    int     i = 0;
    int     j = 0;
    int     s = 0;
    int     h = 0;                  /* cmd switch value */
    int     kp = 0;                 /* cmd line: key pos */
    int     kr = 0;                 /*          key reverse */
    int     ks = 0;                 /*          key size */
    char    kt = 0;                 /*          key data type */
    char    m[6] = "r";             /* mask for fopen */
    char    fi[1024] = "";          /* input file name */
    char    fo[1024] = "";          /* output file name */
    char    v[20] = "";             /* temp */
    int     t1 = 0;                 /* time/secs */
    int     t2 = 0;                 /* time/secs */
    int fn_cpy(char *t,char *x);
    int fn_cmp(char *t,char *x);

    for (i = 1; i < argc; i++)
    {
        h = argv[i][0];
        switch (tolower(h))
        {
            case '<': strncpy(fi,&argv[i][1],1022); break;
            case '>': strncpy(fo,&argv[i][1],1022); break;
            case '-':
#ifdef VMS
            case '/':
#endif
            {
                for (j = 1; j < strlen(argv[i]); j++ ) switch (tolower(argv[i][j]))
                {
                    case 'b': bin = rsz = atoi(&argv[i][j+1]); j = strlen(argv[i]); break;
                    case 'l': lrg = 1; break;
                    case 'k': kys = 1; break;
                    case 'd': dbg = 1; break;
                    case 'i': inf = 1; break;
                    case 'q': qst = 1; break;
                    case 'r': kr  = 1; break;
                    case 't': kt  = argv[i][j+1]; j = strlen(argv[i]); break;
                    case 'u': unq = 1; break;
                    case 'p':
                        if (kp|ks) { kp = ks = kr = kt = fn_makeky(kn,kp,ks,kt,kr); kn++; }
                        kp  = atoi(&argv[i][j+1]);
                        j = strlen(argv[i]);
                        break;
                    case 's':
                        ks  = atoi(&argv[i][j+1]);
                        if (kp|ks) { kp = ks = kr = kt = fn_makeky(kn,kp,ks,kt,kr); kn++; }
                        j = strlen(argv[i]);
                        break;
                    case 'w': Dcd = argv[i][j+1]; j = strlen(argv[i]); break;
                    default: fn_help(); break;
                }
                break;
            }
            default:
            {
                if (kp|ks) { kp = ks = kr = kt = fn_makeky(kn,kp,ks,kt,kr); kn++; }
                if (fi[0] == 0) strncpy(fi,argv[i],1022);
                else
                if (fo[0] == 0) strncpy(fo,argv[i],1022);
            }
        }
    }

    if (bin) strcpy(m,"rb");
    if (fi[0]) {
        if (fn_open(fi,m,&fip)) fn_abort("can't open input",fi);
    } else {
        fip = stdin;
    }
    if (fo[0])
    {
        if (bin)
        {
            strcpy(m,"r+b");
            if (fn_open(fo,m,&fop))
            {
            strcpy(m,"w+b");
            if (fn_open(fo,m,&fop)) fn_abort("can't open output",fo);
            }
        }
        else
        if ((fop = freopen(fo ,"w" ,stdout
#ifdef VMS
            , "rfm=var", "rat=cr", "mrs=0"
#endif
            )) == NULL) fn_abort("can't open output file",fo);
    }

    if (fop == NULL)
        if (bin)
            fn_abort("must have output file for binary","");
        else fop = stdout;

    if (!kn)
        if (fn_makeky(kn,kp,ks,kt,kr))
            fn_abort("can't make key0","");
        else kn++;

    /*
    this is kludgy, here; it ought to be in the init functions, but it has
    to come after all the keys are made:
    */
    if (!kys) for (i = 0; i < kn; i++) ky[i]->tos = ky[i]->pos;

    /* step 1:  read & tag all the records */
    while (!fn_read(&ip,0,tx))
        if (fn_maketg(tn,&ip))
            fn_abort("can't make a tag",tx);
        else
            tn++;

    if (dbg) for (i = 0; i < tn; i++)
        fprintf(stderr,
            "%4d %8p %8p %8d %s\n",i,&tg[i],tg[i],tg[i]->ptr,tg[i]->tag);

    /* step 2:  sort 'em */
    if (tn)
    {
        if (inf)
        {
            t1 = fn_time();
            fprintf(stderr," srt %s\n",VERSION);
        }
        if (qst)
            qsort(tg,tn,sizeof(TG *),fn_compare);
        else
            fn_order(tg,tn);
        if (inf)
        {
            t2 = fn_time();
            fprintf(stderr," %d seconds in sort, %d records\n",t2-t1,tn);
        }
    }

    if (dbg) for (i = 0; i < tn; i++)
        fprintf(stderr,
            "%4d %8p %8p %8d %s\n",i,&tg[i],tg[i],tg[i]->ptr,tg[i]->tag);

    /* step 3:  write the (sorted) records */
    for (rnm = 0; rnm < tn; rnm++)
    {
        fn_read(&tg[rnm]->ptr,(1+lrg+kys),tx);
        if (unq) {
            if (fn_cmp(tx,xx)) {
                fn_write(tx);
                fn_cpy(xx,tx);
            }
        } else {
            fn_write(tx);
        }
    }

    if (fip) fclose(fip);
    if (fop) fclose(fop);

    exit(0);
}

void fn_order(TG **A, int N)
{
    int i = 0;
    int j = 0;
    int h = 0;
    int l = 0;
    int m = 0;
    int s = 0;
    int x = 0;
    int y = 0;
    int z = 0;
    TG *I = NULL;
    TG *L = NULL;
    TG *H = NULL;
    TG *X = NULL;
    TG *Y = NULL;
    TG *Z = NULL;
    int *B = NULL;
    TG **C = NULL;

    B = calloc(N+1,sizeof(int));
    C = calloc(N+1,sizeof(TG *));
    s = 1 + (int)sqrt(N);
    L = H = A[0];
    for (x = 0; x < N; x++) {
        X = A[x];
        m = 1;
        if (fn_compare(&X,&L) < 0) {
            B[x] = l;
            L = X; l = x;
            m = 0;
        }
        if (fn_compare(&H,&X) <= 0) {
            B[h] = x;
            B[x] = 0;
            H = X; h = x;
            m = 0;
        }
        if (m) {
            s = 1 + (int)sqrt(x);
            Y = NULL;
            for (i = 0; i < x; i+=s) {
                I = A[i];
                if ((fn_compare(&I,&X) < 0) && (fn_compare(&I,&Y) > 0)) {
                    Y = I; y = i;
                }
            }
            if (!Y) {
                Y = L; y = l;
            }
            i = y;
            while (m) {
                z = i;
                i = B[i];
                I = A[i];
                if (fn_compare(&I,&X) > 0) {
                    m = 0;
                }
            }
            B[z] = x;
            B[x] = i;
        }
    }
    i = l;
    for (x = 0; x < N; x++) {
        C[x] = A[i];
        i = B[i];
    }
    for (x = 0; x < N; x++) {
        A[x] = C[x];
    }
}

int fn_read(long *p,int w,char *t)
{
    int     i = 0;
    int     m = 0;
    int     r = 0;

    if (w)
    {
        if (w > 1)
        {
            fseek(fip, *p, SEEK_SET);
        }
        else
        {
            if (bin) memcpy(t,tg[rnm]->tag,bin);
            else   { strcpy(t,tg[rnm]->tag); rsz = strlen(t); }
            return(0);
        }
    }
    else *p = ftell(fip);

    m = (bin ? bin : _LREC_);
    for (i = 0; i < m; i++)
    {
        if ((r = fgetc(fip)) == EOF) break;
        t[i] = r;
        if ((!bin) && (r == '\n')) break;
    }
    if (r == EOF)
    {
        if (ferror(fip)) fn_abort("input error","");
        return(1);
    }
    if (!bin) t[++i] = '\0';
    rsz = i;
    /* fprintf(stderr,"r %d %d %d %d\n",t[i-3],t[i-2],t[i-1],t[i]); */
    return(0);
}

int fn_cmp(char *t,char *x)
{
    int     i = 0;

    for (i = 0; i < rsz; i++) {
        if (t[i] > x[i]) return(1);
        if (t[i] < x[i]) return(-1);
    }
    return(0);
}

int fn_cpy(char *t,char *x)
{
    int     i = 0;

    for (i = 0; i < rsz; i++) t[i] = x[i];
    return(0);
}

int fn_write(char *t)
{
    int     i = 0;

    for (i = 0; i < rsz; i++) fputc(t[i],fop);
    if (ferror(fop)) fn_abort("output error","");
    /* fprintf(stderr,"w %d %d %d\n",t[i-3],t[i-2],t[i-1]); */
    return(0);
}

int fn_open(char *fn, char *m, FILE **fp)
{
    if (!fn[0]) return(1);
    if ((*fp = fopen(fn,m
#ifdef VMS
	,"shr=upd"
#endif
        )) == NULL) return(1);
    return(0);
}

/*
    fn_compare has to behave like strcmp for qsort to work correctly; it
    will pass control to the array of specialized functions, then give
    back a value of -1, 0, or 1. 
*/
int fn_compare(const void *p1, const void *p2)
{
    static char t1[_LREC_];
    static char t2[_LREC_];
    TG **tp1;
    TG **tp2;
    TG *tg1;
    TG *tg2;
    int i = 0;
    int j = 0;
    int k = 0;
    /* each key type has its own comparison function in this array: */
    int (*fn_cmp[])() = 
        { fn_strcmp             /* K_STR 0 */
        , fn_intcmp             /* K_INT 1 */
        , fn_dt6cmp             /* K_DT6 2 */
        , fn_mdycmp             /* K_MDY 3 */
        , fn_numcmp             /* K_NUM 4 */
        , fn_lexcmp };          /* K_LEX 5 */

    if (p1 == p2) return(0);
    if (p1 == NULL) return(-1);
    if (p2 == NULL) return(1);
    tp1 = (TG **)p1;
    tp2 = (TG **)p2;
    tg1 = *tp1;
    tg2 = *tp2;
    if (tg1 == NULL) return(-1);
    if (tg2 == NULL) return(1);
    if (tg1->tag == NULL) return(-1);
    if (tg2->tag == NULL) return(1);
    if (dbg) {
        fprintf(stderr,"|%s|<==>\n",tg1->tag);
        fprintf(stderr,"|%s| ",tg2->tag);
    }

    /*
    if it's a large file, (-l switch) we don't have the 'tags' in
    memory, so we have to read them in to compare:
    */
    if (lrg)
    {
        tg1->tag = t1;
        tg2->tag = t2;
        fn_read(&tg1->ptr,2,t1);
        fn_read(&tg2->ptr,2,t2);
    }
/*
    if (!ky[0]->siz)
    {
        if (bin)
            k = memcmp(&tg1->tag[ky[0]->tos]
                      ,&tg2->tag[ky[0]->tos]
                            ,bin-ky[0]->tos);
        else
            k = strcmp(&tg1->tag[ky[0]->tos]
                      ,&tg2->tag[ky[0]->tos]);
        if (dbg) fprintf(stderr,"%d\n",k);
        return(k *               ky[0]->rev);
    }
    else
*/
    for (i = 0; i < kn; i++)
    {
        if (dbg) fprintf(stderr,"k %d: ",i);
        k = (*fn_cmp[ky[i]->typ])
               (&tg1->tag[ky[i]->tos]
               ,&tg2->tag[ky[i]->tos]
                         ,ky[i]->siz);
        if (dbg) fprintf(stderr," res: %d\n",k);
        if (k) return(k * ky[i]->rev);
    }
    return(0);
}

int fn_strcmp(char *d1, char *d2, int n)
{
    if (n <= 0) n = (strlen(d1) < strlen(d2) ? strlen(d1) : strlen(d2));
    return(strncmp(d1,d2,n));
}

/*
    fn_lexcmp is not an 'official' lexical comparison; it simply skips
    anything that isn't a letter (alpha), and ignores case;
    not optimized for speed, at all. 
*/
int fn_lexcmp(char *d1, char *d2, int n)
{
    int i = 0;
    int j = 0;
    int k = 0;

    if (n <= 0) n = (strlen(d1) < strlen(d2) ? strlen(d1) : strlen(d2));

    while (!k)
    {
        while ((i < n) && !isalpha(d1[i])) i++;
        while ((j < n) && !isalpha(d2[j])) j++;
        if ((i < n) && (j < n))
        {
            if ((d1[i] & 31) > (d2[j] & 31)) k = 1;
            else
            if ((d1[i] & 31) < (d2[j] & 31)) k = -1;
            else { i++; j++; }
        }
        else if (i < j) k = 1;
        else if (i > j) k = -1;
        else break;
    }
    return(k);
}

/*
    fn_numcmp pretends the key string is a number;
    has to have a length to work right.
*/
int fn_numcmp(char *d1, char *d2, int n)
{
    double n1 = 0.;
    double n2 = 0.;
    char c1[260] = "";
    char c2[260] = "";

    strncpy(c1,d1,n%256);
    strncpy(c2,d2,n%256);
    n1 = atof(c1);
    n2 = atof(c2);
    if (n1 > n2) return(1);
    if (n1 < n2) return(-1);
    return(0);
}

/*
    fn_mdycmp does a 2-byte compare of the year, reversing the sign if
    it's a date in the 90's compared to a date in the 00's; if the year
    matches, we go on to check the 'mmdd'. 
*/
int fn_mdycmp(char *d1, char *d2, int n)
{
    int k = 0;

    k = strncmp(&d1[4],&d2[4],2);
    if (k)
    {
        if (((d1[4] >= Dcd) && (d2[4] <  Dcd))
        ||  ((d1[4] <  Dcd) && (d2[4] >= Dcd))) k = -k;
    }
    else k = strncmp(d1,d2,4);
    return(k);
}

/*
    fn_dt6cmp is a straight 6-byte compare, unless it's a date in the 90's
    compared to a date in the 00's, in which case we reverse the sign.
*/
int fn_dt6cmp(char *d1, char *d2, int n)
{
    int k = 0;

    k = strncmp(d1,d2,6);
    if (k)
    {
        if (((d1[0] >= Dcd) && (d2[0] <  Dcd))
        ||  ((d1[0] <  Dcd) && (d2[0] >= Dcd))) k = -k;
    }
    return(k);
}

/* straight integer comparison */
int fn_intcmp(void *i1, void *i2, int n)
{
    if (((int *)i1) < (int *)i2) return(-1);
    if (((int *)i1) > (int *)i2) return(1);
    return(0);
}

/*

    make a 'tag'; we'll either save the whole record, and its byte
    address, or just the address, or the address and a key; if we save the
    record or the key, we'll allocate space for it on the fly - not fast,
    but reliable, and keeps memory usage to the minimum; we keep expanding
    the pointer array, since we don't know how many elements we'll need (1
    per record of input). 

*/
int fn_maketg(int tn,long *p)
{
    int     i = 0;
    int     j = 0;
    int     k = 0;

    if ((tg = (TG **)realloc(tg,sizeof(TG *)*(tn+1))) == NULL) return(1);
    if ((tg[tn] = (TG *)calloc(1,sizeof(TG))) == NULL) return(1);
    tg[tn]->ptr = *p;
    *p = 0;
    if (kys)
    {
        if ((tg[tn]->tag = (char *)calloc(1,ksz+1)) == NULL) return(1);
        for (i = 0; i < kn; i++)
        {
            k = ((ky[i]->siz <= 0) ? rsz - ky[i]->pos : ky[i]->siz);
            memcpy( &tg[tn]->tag[j]
                   ,&tx[ky[i]->pos]
                       ,k);
            j +=        k;
        }
    }
    else
    if (!lrg)
    {
        if ((tg[tn]->tag = (char *)calloc(1,rsz+1)) == NULL) return(1);
        memcpy(tg[tn]->tag,tx,rsz);
    }
    return(0);
}

/*
    build a key; reallocate the array of key pointers as we go; no limit
    on number or type of keys; each key gets a position (default 0), an
    offset within the whole key, a size (might be 0, but only if no keys
    were requested), a direction sign (-1=reverse, 1=normal), and a type
    (default 0/string). 
*/
int fn_makeky(int kn,int kp,int ks,char kt,int kr)
{
    if ((ky = (KY **)realloc(ky,sizeof(KY *)*(kn+1))) == NULL) return(1);
    if ((ky[kn] = (KY *)calloc(1,sizeof(KY))) == NULL) return(1);
    ky[kn]->rev = ((kr) ? -1 : 1);
    ky[kn]->siz = ks;
    ky[kn]->pos = kp;
    ky[kn]->tos = ksz; ksz += ks;
    switch (tolower(kt))
    {
        case 'i':   ky[kn]->typ = K_INT; break;
        case 'l':   ky[kn]->typ = K_LEX; break;
        case 'm':   ky[kn]->typ = K_MDY; break;
        case 'n':   ky[kn]->typ = K_NUM; break;
        case 'd':   ky[kn]->typ = K_DT6; break;
        case 's':
        default:    ky[kn]->typ = K_STR; break;
    }
    return(0);
}

int fn_time()
{
#include <time.h>

    time_t      t;
    struct tm  *ts;

    time(&t);
    ts = localtime(&t);
    fprintf(stderr," %02d:%02d:%02d "
        ,ts->tm_hour
        ,ts->tm_min
        ,ts->tm_sec);
    return(t);
}

void fn_abort(char *msg, char *vbl)
{
    fprintf(stderr,"\n!! %s %s\n",msg,vbl);
    exit(2);
}

void fn_help()
{
    fprintf(stderr," SRT (sort utility) syntax:\n");
    fprintf(stderr,"  srt -<switches> infile.dat [outfile]\n\n");
    fprintf(stderr,"  switches:\n");
    fprintf(stderr,"  -b# : Binary file; define output file before sort\n");
    fprintf(stderr,"        # = bytes/record, must not be 0\n");
    fprintf(stderr,"  -d  : Debug, show pointers before & after qsort\n");
    fprintf(stderr,"  -i  : Info, show stats\n");
    fprintf(stderr,"  -k  : Key sort, saves memory, faster than -l\n");
    fprintf(stderr,"  -l  : Large file, work on the disk (slow)\n");
    fprintf(stderr,"  -q  : use Qsort, instead of Ordersort\n");
    fprintf(stderr,"  -w# : change Window decade fr 90, for mmddyy\n");
    fprintf(stderr,"  keys:\n");
    fprintf(stderr,"  -p  : key Position (offset from 0)\n");
    fprintf(stderr,"  -r  : Reverse key (descending sort)\n");
    fprintf(stderr,"  -ts : key data Type, s = (one of these)\n");
    fprintf(stderr,"        s(string),          i(int)\n");
    fprintf(stderr,"        d(date, yymmdd)     m(mmddyy)\n");
    fprintf(stderr,"        l(lexical, alphabetic sort)\n");
    fprintf(stderr,"        n(numeric field, unjustified)\n");
    fprintf(stderr,"  -s  : key Size (width in bytes/characters)\n");
    fprintf(stderr,"  -u  : like uniq, remove duplicate records\n\n");
    fprintf(stderr,"  Specify keys with Pos switch first, Size switch last for\n");
    fprintf(stderr,"  each key, e.g.: srt in out -p10 -td -s6 -p0 -tl -s10\n");
    fprintf(stderr,"  which specifies two keys, a date and a lexical.\n\n");
    fprintf(stderr,"  To make srt on this platform:\n    %s\n\n",MAKE_SRT);
    fprintf(stderr,"  %s %s -%s\n",VERSION,BUILD,AUTHOR);
    exit(0);
}
